// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Prog4D Client Wrapper                                            ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


import java.io.*;
import mpi_graphics.*;


/**
  * This class is a wrapper to Prog4D client functions.
  *
  * @short   Prog4D Client
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
public class Prog4DClient
{
   private mpi_graphics.PM_Anim_Bonn       animator       = null;
   private bonn.infoIV.v3d2.rtp.RTPSession rtpsession     = null;
   private boolean                         playing        = false;
   private long                            bandwidthLimit;

   private String location;

   private int RTPMaxPayload = 1012;
   private int BaseBps    = (2000 + 3 * (20 + 12 + 8 + 4));
   private int ExtBps     = (RTPMaxPayload + (20 + 12 + 8));
   private int MaximumBps = ((30 * BaseBps) + ((20000 / 8)) * ExtBps);
   private int StartBps   = ((30 * BaseBps) + (1000 / 8) * ExtBps);


   // ###### Constructor ####################################################
   /**
     * Constructor.
     *
     * @param videoLocation Location of video input or null for default.
     */
   public Prog4DClient(String videoLocation) {
      if(videoLocation == null)
         location = "localhost";
      else
         location = videoLocation;
   }


   // ###### Start playing ##################################################
   /**
     * Start playing.
     */
   synchronized public void play() {
      if(playing == false) {
         bandwidthLimit = (20 * 1000000) / 8;
         try {
            rtpsession = new bonn.infoIV.v3d2.rtp.RTPSession(
                               java.net.InetAddress.getByName(location),
                               6000,5000);
            animator = new PM_Anim_Bonn(rtpsession);
            rtpsession.startSession(animator);
            animator.setBpsLimit((int)bandwidthLimit * 8);
            animator.start();
            playing = true;            
         } catch(Exception e) {
            System.err.println("ERROR: Prog4DClient::play() - Prog4D exception: " + e);
            stop();
         }
      }
   }


   // ###### Stop playing ###################################################
   synchronized public void stop() {
      if(animator != null) {
         animator.stop();
         rtpsession.shutdown();
         animator.shutdown();
      }
      else if(rtpsession != null) {
         rtpsession.shutdown();
      }
      animator   = null;
      rtpsession = null;
      playing    = false;
   }


   // ###### Get transport info constants ###################################
   /**
     * Get TransportInfo constants.
     *
     * @param transportInfo TransportInfo.
     */
   synchronized public void getTransportInfoConstants(TransportInfo transportInfo) {
      // ====== Initialize wanted values ====================================
      transportInfo.MinBytesPerSecond[0]   = BaseBps;
      transportInfo.MaxBytesPerSecond[0]   = BaseBps;
      transportInfo.MinPacketsPerSecond[0] = -1;
      transportInfo.MaxPacketsPerSecond[0] = -1;
      transportInfo.WantedMaxLossRate[0]   = 0;

      transportInfo.MinBytesPerSecond[1]   = ExtBps;
      transportInfo.MaxBytesPerSecond[1]   = MaximumBps - BaseBps;
      transportInfo.MinPacketsPerSecond[1] = -1;
      transportInfo.MaxPacketsPerSecond[1] = -1;
      transportInfo.WantedMaxLossRate[1]   = 25;

      for(int i = 2;i < TransportInfoLevel.MaxQualityLayers;i++) {
         transportInfo.MinBytesPerSecond[i]   = 0;
         transportInfo.MaxBytesPerSecond[i]   = 0;
         transportInfo.MinPacketsPerSecond[i] = 0;
         transportInfo.MaxPacketsPerSecond[i] = 0;
         transportInfo.WantedMaxLossRate[i]   = 0;
      }

      transportInfo.MinFramesPerSecond      = 1;
      transportInfo.MaxFramesPerSecond      = 30;
      transportInfo.WantedMaxTransferDelay  = 1000 * 16;
      transportInfo.StartFramesPerSecond    = 1;

      transportInfo.Flags         = TransportInfo.TIF_HasDynamic;
      transportInfo.QualityLevels = 1;
      transportInfo.QualityLayers = 2;
      

      // ====== Initialize level and layer ==================================
      transportInfo.QualityLevel[0].BytesPerSecondScale   = 65536;
      transportInfo.QualityLevel[0].PacketsPerSecondScale = 65536;
      transportInfo.QualityLevel[0].FramesPerSecondScale  = 65536;
      transportInfo.QualityLevel[0].FramesPerSecond       = 1;
      transportInfo.QualityLevel[0].MaxTransferDelay      = 16 * 1000;
      transportInfo.QualityLevel[0].Quality               = 255;
      transportInfo.QualityLevel[0].LevelUp               = 255;
      transportInfo.QualityLevel[0].LevelDown             = 255;
      transportInfo.QualityLevel[0].QualityLayers         = 2;

      transportInfo.QualityLevel[0].QualityLayer[0].BytesPerSecond   = BaseBps;
      transportInfo.QualityLevel[0].QualityLayer[0].PacketsPerSecond = -1;
      transportInfo.QualityLevel[0].QualityLayer[0].FrameSize        = BaseBps;
      transportInfo.QualityLevel[0].QualityLayer[0].MaxLossRate      = 0;
      transportInfo.QualityLevel[0].QualityLayer[0].Flags            = TransportInfoLayer.TILF_Dynamic;

      transportInfo.QualityLevel[0].QualityLayer[1].BytesPerSecond   = ExtBps;
      transportInfo.QualityLevel[0].QualityLayer[1].PacketsPerSecond = -1;
      transportInfo.QualityLevel[0].QualityLayer[1].FrameSize        = ExtBps;
      transportInfo.QualityLevel[0].QualityLayer[1].MaxLossRate      = 25;
      transportInfo.QualityLevel[0].QualityLayer[1].Flags            = TransportInfoLayer.TILF_Dynamic;


      // ====== Initialize current settings =================================
      transportInfo.CurrentSetting.BytesPerSecondScale   = 65536;
      transportInfo.CurrentSetting.PacketsPerSecondScale = 65536;
      transportInfo.CurrentSetting.FramesPerSecondScale  = 65536;
      transportInfo.CurrentSetting.FramesPerSecond       = 1;
      transportInfo.CurrentSetting.MaxTransferDelay      = 16 * 1000;
      transportInfo.CurrentSetting.Quality               = 255;
      transportInfo.CurrentSetting.LevelUp               = 255;
      transportInfo.CurrentSetting.LevelDown             = 255;
      transportInfo.CurrentSetting.QualityLayers         = 2;

      transportInfo.CurrentSetting.QualityLayer[0].FrameSize   = BaseBps;
      transportInfo.CurrentSetting.QualityLayer[0].MaxLossRate = 0;
      transportInfo.CurrentSetting.QualityLayer[0].Flags       = TransportInfoLayer.TILF_Static;

      transportInfo.CurrentSetting.QualityLayer[1].FrameSize   = ExtBps;
      transportInfo.CurrentSetting.QualityLayer[1].MaxLossRate = 25;
      transportInfo.CurrentSetting.QualityLayer[1].Flags       = TransportInfoLayer.TILF_Dynamic;


      if(animator != null) {
         transportInfo.CurrentSetting.FramesPerSecond = animator.getFps();
         transportInfo.CurrentSetting.QualityLayer[0].BytesPerSecond = BaseBps * animator.getFps();
         if(transportInfo.CurrentSetting.QualityLayer[0].BytesPerSecond > bandwidthLimit) {
            transportInfo.CurrentSetting.QualityLayer[0].BytesPerSecond = 0;
         }
         transportInfo.CurrentSetting.QualityLayer[1].BytesPerSecond = (animator.getBps() / 8) - (BaseBps * animator.getFps());
         if(transportInfo.CurrentSetting.QualityLayer[1].BytesPerSecond > bandwidthLimit - (BaseBps * animator.getFps())) {
            transportInfo.CurrentSetting.QualityLayer[1].BytesPerSecond = 0;
         }
         transportInfo.CurrentSetting.QualityLayer[0].PacketsPerSecond = -1;
         transportInfo.CurrentSetting.QualityLayer[1].PacketsPerSecond = -1;
      }
      else {
         transportInfo.CurrentSetting.QualityLayer[0].BytesPerSecond   = 0;
         transportInfo.CurrentSetting.QualityLayer[1].BytesPerSecond   = 0;
         transportInfo.CurrentSetting.QualityLayer[0].PacketsPerSecond = 0;
         transportInfo.CurrentSetting.QualityLayer[1].PacketsPerSecond = 0;
      }
   }


   // ###### Get transport info #############################################
   /**
     * Get TransportInfo for current setting.
     *
     * @param transportInfo TransportInfo.
     */
   synchronized public void getTransportInfo(TransportInfo transportInfo) {
      getTransportInfoConstants(transportInfo);

      if(animator != null) {
         transportInfo.WantedBytesPerSecond[0]   = 2000 * animator.getFps();
         transportInfo.WantedBytesPerSecond[1]   = (animator.getBps() / 8) - 2000;
         transportInfo.WantedPacketsPerSecond[0] = -1;
         transportInfo.WantedPacketsPerSecond[1] = -1;
         transportInfo.WantedFramesPerSecond     = animator.getFps();
      }
      else {
         transportInfo.WantedBytesPerSecond[0]   = transportInfo.MinBytesPerSecond[0];
         transportInfo.WantedBytesPerSecond[1]   = transportInfo.MinBytesPerSecond[1];
         transportInfo.WantedPacketsPerSecond[0] = transportInfo.MinPacketsPerSecond[0];
         transportInfo.WantedPacketsPerSecond[1] = transportInfo.MinPacketsPerSecond[1];
         transportInfo.WantedFramesPerSecond     = transportInfo.MinFramesPerSecond;
      }
   }


   // ###### Set transport info #############################################
   /**
     * Set TransportInfo.
     *
     * @param transportInfo TransportInfo.
     */
   synchronized public void setTransportInfo(TransportInfo transportInfo) {
      getTransportInfoConstants(transportInfo);
      long bytesPerSecond = transportInfo.WantedBytesPerSecond[0] +
                            transportInfo.WantedBytesPerSecond[1];

      long bps = bytesPerSecond * 8;
      int fps  = transportInfo.WantedFramesPerSecond;

      if(bps > transportInfo.MaxBytesPerSecond[0] + transportInfo.MaxBytesPerSecond[1])
         bps = transportInfo.MaxBytesPerSecond[0] + transportInfo.MaxBytesPerSecond[1];
      if(bps < transportInfo.MinBytesPerSecond[0] + transportInfo.MinBytesPerSecond[1])
         bps = transportInfo.MinBytesPerSecond[0] + transportInfo.MinBytesPerSecond[1];

      if(fps > transportInfo.MaxFramesPerSecond)
         fps = transportInfo.MaxFramesPerSecond;
      if(fps < transportInfo.MinFramesPerSecond)
         fps = transportInfo.MinFramesPerSecond;

      transportInfo.WantedBytesPerSecond[0]   = 2000 * fps;
      transportInfo.WantedBytesPerSecond[1]   = (bps / 8) - (2000 * fps);
      transportInfo.WantedPacketsPerSecond[0] = -1;
      transportInfo.WantedPacketsPerSecond[1] = -1;
      transportInfo.WantedFramesPerSecond     = fps;

      if(animator != null) {
         animator.setBps((int)bps);
         animator.setFps(fps);
      }
   }


   // ###### Set bandwidth ##################################################
   /**
     * Set bandwidth limit.
     *
     * @param bandwidthLimit Bandwidth limit.
     */
   void setBandwidthLimit(long bandwidth) {
      bandwidthLimit = bandwidth;
      if(animator != null)
         animator.setBpsLimit((int)bandwidthLimit * 8);
   }


   // ###### Main program ###################################################
   /**
     * Main program.
     * This will start a test client.
     */
   public static void main(String[] args) {
      String location = null;
      if(args.length > 0) {
         location = args[0];
      }

      Prog4DClient client = new Prog4DClient(location);
      TransportInfo ti = new TransportInfo();

      client.play();

      for(;;) {
//         System.out.println("============ TransportInfo ============");
//         client.getTransportInfo(ti);
//         ti.print();
         try {
            Thread.sleep(5000);
         } catch(Exception e) { }
      }
   }
}
