//########################################################
//##                                                    ##
//##                 QoS Management                     ##
//##                 ==============                     ##
//##                                                    ##
//## QoS-Manager                                        ##
//##                                                    ##
//## Version 1.0                                        ##
//## Date  : March 2000                                 ##
//## Author: Simon Vey (vey@cs.uni-bonn.de)             ##
//##                                                    ##
//########################################################

#ifndef QOS_MANAGER
#define QOS_MANAGER


#include "qosglobals.h"
#include "../timedthread.h"
#include <vector>
#include "session.h"
#include "layeredstream.h"
#include "../ResvMod/resvmod.h"
#include "qosmanagerinterface.h"
#include "../tools.h"
#include <fstream>
#include "../extendedtransportinfo.h"
#include "../internetaddress.h"
#include "../roundtriptimepinger.h"

// ??????????????????
#include "../rtcppacket.h"


namespace Coral {

/**
 * This class implements the QoS-Manager. 
 *
 * @short   Session
 * @author  Simon Vey (vey@cs.uni-bonn.de)
 * @version 1.0
 * @see
 */

class QoSManager : public TimedThread {
 public:
  /**
   * Constructor.
   */
  QoSManager();

  ~QoSManager();

  /**
   * Creates a new Session to the given destination address
   * 
   * @return Session ID
   */
  card32 newSession(PortableAddress dest);

  /**
   * Close the session with the given ID
   *
   * @param id Session identifier
   * @return true, if session is closed, false otherwise.
   */
  bool closeSession(card32 id);

  /**
   * Sets the weight of a session.
   *
   * @param sessionID ID of the session.
   * @param weight new weight of the session.
   * @return false, if there exists no session with this ID.
   */
  bool setSessionWeight(card32 sessionID, card32 weight);

  /**
   * Returns the weight of the session.
   *
   * @param sessionID ID of the session.
   * @return Weight of the session.
   */
  card32 getSessionWeight(card32 sessionID);

  /**
   * Adds a new stream to a specified session.
   *
   * @param sessionID Session, to wich the flow schould be added.
   * @param ti TransportInfo of the flow.
   * @param interface Interface for application callback.
   * @return Identifier of the new flow.
   */
  card32 newStream(card32 ID, ExtendedTransportInfo ti, QoSManagerInterface* interface);

  /**
   * Closes a flow in the specified session.
   *
   * @param sessionID Session identifier.
   * @param flowID ID of the flow to close.
   * @return true, if flow has been closed, false otherwise.
   */
  bool closeStream(card32 ID);

  /**
   * creates a new stream-ID.
   * This ID's must be given to newStream()
   */
  int getID();

  /**
   * Sets the loss-fraction (0<=fraction<=1) of a given layer of a given stream.
   * If this fraction is bigger than allowed for this layer, the stream will be
   * scaled down.
   */
  void setFractionLost(int flowID, double fraction, card8 layerNumber);


  // ???????????????????
  void setReport(int flowID, const RTCPReceptionReportBlock* report, card8 layerNumber);
  // ???????????????????
  

  /**
   * Updates the TransportInfo for a stream.
   */
  int updateStream(const int fid, const ExtendedTransportInfo& ti);

  int setWeight(int flowID, short weight, bool newCalc);


  /**
   * Returns the bandwidth of the class.
   */
  card64 getBandwidthOfClass(card8 class_);

  /**
   * Returns the total bandwidth.
   */
  card64 getTotalBandwidth();

  /**
   * Returns the free bandwidth of the class.
   */
  card64 getFreeBandwidthOfClass(card8 class_);

  /**
   * sets the free bandwidth of a class.
   */
  void   setFreeBandwidthOfClass(card8 class_, card32 bandwidth);

  /**
   * returns the calculation interval.
   */
  card32 getCalculationInterval();

  /**
   * Sets the calculation intervall.
   */
  void   setClaculationInterval(card32 time);

  //  Session* getSession(card32 id);
  
  /**
   * returns the number of priority classes.
   */
  card8  getNumPrios();
  card8  getMaxLayerNumber();

 private:
  
  card64 totalBandwidth;             // total bandwidth <---------------Einheiten !!!!!

  vector<Session*> session;          // sessions to manage
  
  // const static card8 numPrios = 16;  // number of priority classes

  const static card8 maxLayerNumber = 4;

  card64 prio[NUM_PRIOS];            // available bandwidth of each class

  card64 freeInPrio[NUM_PRIOS];      // current free bandwidth of each class

  RESV_Modul ResvMod;                // reservation modul

  ofstream* outfile;                 // for log-file

  ofstream* gnuscript;
  ofstream* daten;
  ofstream* scaleOut;
  ofstream* qualityOut;

  //card32 sessionOfFlow[20];

  RoundTripTimePinger* rttp;

  card32 calculationInterval;        // in seconds

  // card32 minScaleInterval;

  void timerEvent();               // called once every second

  void initRTTPinger();

  // adds a new stream to the priority classes 
  void addStreamToClasses(LayeredStream* newStreamP, card32 sessionIndex, int level);

  //  bool classFound(Flow* newFlowP, card8 level , bool* classOk, card64 bw);
  /**
   * Search a session to scale down.
   *
   * @return true, if a session is scaled down, false otherwise.
   */
  //  bool freeBandwidth();

  /**
   * calculates new bandwidth distribution.
   * (called from 'timerEvent' in a fixed interval)
   */
  void calculateNewMapping();

  // callback-function (calls resv-modul for setting the traffic-class,
  // sets the values in the transportinfo and calls the appliction-callback-function)
  void applicationCallback();

  void writeLogInfo();

  card32 scaleCount;
};

#include "sortstruct.h"

}

#include "qosmanager.icc"
#endif
