// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Transport Info                                                   ####
// ####                                                                  ####
// #### Version 1.00, February 25, 2000                                  ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://titan.cs.uni-bonn.de/~dreibhol ####
// ####                                                                  ####
// ##########################################################################


#ifndef TRANSPORTINFO2_H
#define TRANSPORTINFO2_H


#include "../system.h"
#include "../range.h"
#include "level.h"


namespace Coral {


/**
  * This class keeps information on an encoding's transport of data,
  * e.g. bytes, frames and packets per second.
  *
  * @short   Transport Information
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  *
  * @see EncoderInterface
  * @see DecoderInterface
  */
class TransportInfo2
{
   // ====== Constructor ====================================================
   public:
   /**
     * Constructor.
     */
   TransportInfo2();

   // ====== Byte order translation =========================================
   /**
     * Translate byte order.
     */
   void translate();

   // ====== Reset ==========================================================
   /**
     * Reset.
     */
   void reset();


   // ====== Values =========================================================
   public:
   // ====== Minimum, maximum and wanted settings ===========================
   /**
     * Bytes per second: Minimum, maximum and current quality setting's value.
     * These values contain base data + extension data!
     */
   Range<card64> WantedBytesPerSecond;

   /**
     * Wanted bytes per second for extension data.
     */
   card64 WantedExtensionBytesPerSecond;

   /**
     * Frames per second: Minimum, maximum and current quality setting's value.
     */
   Range<card32> WantedFramesPerSecond;

   /**
     * Packets per second: Minimum, maximum and current quality setting's value.
     * These values contain base data + extension data!
     */
   Range<card32> WantedPacketsPerSecond;

   /**
     * Wanted packets per second for extension data.
     */
   card32 WantedExtensionPacketsPerSecond;

   /**
     * Wanted maximum transfer delay in 1/16 milliseconds for base data.
     */
   card32 WantedMaxTransferDelay;

   /**
     * Wanted maximum transfer delay in 1/16 milliseconds for extension data.
     */
   card32 WantedMaxExtensionTransferDelay;

   
   // ====== Flags ==========================================================
   /**
     * TransportInfo flags.
     */
   card8 Flags;

   /**
     * TransportInfo flags enumeration.
     */
   enum TransportInfoFlags {
      TIF_None            = 0,
      TIF_LayeredEncoding = (1 << 0)
   };


   // ====== Traffic class setting ==========================================
   /**
     * Traffic class/TOS setting.
     * This value can be given to the sender to set traffic class for outgoing
     * packets.
     */
   card8 TrafficClass;

   /**
     * Extension Traffic class/TOS setting.
     * This value can be given to the sender to set traffic class for outgoing
     * packets.
     */
   card8 ExtensionTrafficClass;

   // SV===================================================================SV
   /**
    * Start value for the frames per second of a stream.
    * This value can be scaled by each level.
    */
   card8 start_fps;

   /**
    * Weight of stream against other stream types
    */
   card8 bw_weight;
   
   // SV===================================================================SV


   // ====== Quality level constants ========================================
   /**
     * Constant for maximum number of quality levels.
     */
   const static card8 MaxQualityLevels = 16;

   /**
     * Number of quality levels stored in QualityLevel array.
     * This is a constant set by the encoding.
     */
   card8 QualityLevels;



   /**
     * Quality level transport information array.
     * All values are *constants* of the encoding.
     */
   Level QualityLevel[MaxQualityLevels];


   // ====== Current quality setting ========================================
   /**
     * Current setting's transport information.
     * The encoding will store current setting here.
     */
   card8 CurrentLevel;
};


/**
  * << operator.
  */
//ostream& operator<<(ostream& os, const TransportInfo& ti);


}


#endif
