// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Advanced Audio Packet Implementation                             ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "advancedaudiopacket.h"
#include "tools.h"


namespace Coral {


// ###### Encoding name #####################################################
const char AdvancedAudioPacket::AdvancedAudioTypeName[] =
   "Advanced Audio Encoding";


// ###### Constructor #######################################################
AdvancedAudioPacket::AdvancedAudioPacket()
{
   reset();
}


// ###### Reset #############################################################
void AdvancedAudioPacket::reset()
{
   FormatID     = AdvancedAudioFormatID;
   Position     = 0;
   MaxPosition  = 0;
   SamplingRate = 0;
   Channels     = 0;
   Bits         = 0;
   ErrorCode    = 0;
   Flags        = 0;
   Fragment     = 0;
}


// ###### Translate byte order ##############################################
void AdvancedAudioPacket::translate()
{
   FormatID     = translate32(FormatID);
   SamplingRate = translate16(SamplingRate);
   Position     = translate64(Position);
   MaxPosition  = translate64(MaxPosition);
   Fragment     = translate16(Fragment);
}


// ###### Get TransportInfo constants #######################################
void AdvancedAudioPacket::getTransportInfoConstants(
        TransportInfo&               transportInfo,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& lastTransmissionQuality,
        const bool                   calculateLevels)
{
   // ====== Calculate minimum and maximum values ===========================
   const card32 minPPS = calculatePacketsPerSecond(
                            AudioQuality::LowestQuality.getBytesPerSecond(),
                            AdvancedAudioFramesPerSecond,
                            maxPacketSize,sizeof(AdvancedAudioPacket) + headerSize);
   const card32 maxPPS = calculatePacketsPerSecond(
                            AudioQuality::HighestQuality.getBytesPerSecond() / 4,
                            AdvancedAudioFramesPerSecond,
                            maxPacketSize,sizeof(AdvancedAudioPacket) + headerSize) +
                         AdvancedAudioMediaInfoPacketsPerSecond;
   const card64 maxBPSn = calculateBytesPerSecond(
                             AudioQuality::HighestQuality.getBytesPerSecond() / 4,
                             AdvancedAudioFramesPerSecond,
                             maxPacketSize,sizeof(AdvancedAudioPacket) + headerSize);
   const card64 minBPS0 = calculateBytesPerSecond(
                             AudioQuality::LowestQuality.getBytesPerSecond(),
                             AdvancedAudioFramesPerSecond,
                             maxPacketSize,sizeof(AdvancedAudioPacket) + headerSize) +
                         (AdvancedAudioMediaInfoPacketsPerSecond *
                             (sizeof(AdvancedAudioPacket) + headerSize + sizeof(MediaInfo)));
   const card64 maxBPS0 = calculateBytesPerSecond(
                              AudioQuality::HighestQuality.getBytesPerSecond() / 4,
                              AdvancedAudioFramesPerSecond,
                              maxPacketSize,sizeof(AdvancedAudioPacket) + headerSize) +
                         (AdvancedAudioMediaInfoPacketsPerSecond *
                             (sizeof(AdvancedAudioPacket) + headerSize + sizeof(MediaInfo)));

   // ====== Calculate quality level array ==================================
   transportInfo.WantedBytesPerSecond[0].setLimits(
      minBPS0,
      maxBPS0);
   transportInfo.WantedBytesPerSecond[1].setLimits(0,maxBPSn);
   transportInfo.WantedBytesPerSecond[2].setLimits(0,2 * maxBPSn);
   transportInfo.WantedPacketsPerSecond[0].setLimits(
      minPPS + AdvancedAudioMediaInfoPacketsPerSecond,
      maxPPS + AdvancedAudioMediaInfoPacketsPerSecond);
   transportInfo.WantedPacketsPerSecond[1].setLimits(0,maxPPS);
   transportInfo.WantedPacketsPerSecond[2].setLimits(0,2 * maxPPS);

   transportInfo.WantedMaxLossRate[0] = 0;
   transportInfo.WantedMaxLossRate[1] = 15;
   transportInfo.WantedMaxLossRate[2] = 30;
   for(cardinal i = 3;i < TransportInfoLevel::MaxQualityLayers;i++) {
      transportInfo.WantedBytesPerSecond[i].init(0,0,0);
      transportInfo.WantedPacketsPerSecond[i].init(0,0,0);
      transportInfo.WantedMaxLossRate[i] = 255;
   }
   transportInfo.WantedFramesPerSecond.init(AdvancedAudioFramesPerSecond,
                                             AdvancedAudioFramesPerSecond,
                                             AdvancedAudioFramesPerSecond);
   transportInfo.WantedMaxTransferDelay = AdvancedAudioMaxTransferDelay;

   transportInfo.Flags                = TransportInfo::TIF_None;
   transportInfo.QualityLevels        = AdvancedAudioQualityLevels;
   transportInfo.QualityLayers        = 3;
   transportInfo.StartFramesPerSecond = AdvancedAudioFramesPerSecond;

   // ====== Calculate all quality levels ===================================
   if(calculateLevels == true) {
      AudioQuality quality = AudioQuality::LowestQuality;
      for(cardinal i = 0;i < AdvancedAudioQualityLevels;i++) {
         calculateLevelForQuality(transportInfo.QualityLevel[i],
                                  headerSize, maxPacketSize, quality);

         transportInfo.QualityLevel[i].LevelUp =
            (i < AdvancedAudioQualityLevels - 1) ? (i + 1) : 0xff;
         transportInfo.QualityLevel[i].LevelDown =
            (i > 0) ? (i - 1) : 0xff;

         quality++;
      }
   }

   // ====== Calculate current setting ======================================
   calculateLevelForQuality(transportInfo.CurrentSetting,
                            headerSize, maxPacketSize,
                            lastTransmissionQuality);
   transportInfo.CurrentSetting.LevelUp   = AdvancedAudioPacket::AdvancedAudioQualityLevels - 1;
   transportInfo.CurrentSetting.LevelDown = 0;
}


// ###### Get TransportInfo #################################################
void AdvancedAudioPacket::getTransportInfo(
        TransportInfo&               transportInfo,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& qualitySetting,
        const AudioQualityInterface& lastTransmissionQuality,
        const bool                   calculateLevels)
{
   // ====== Get constants ==================================================
   getTransportInfoConstants(transportInfo, headerSize,
                             maxPacketSize, lastTransmissionQuality,
                             calculateLevels);

   // ====== Get wanted values ==============================================
   TransportInfoLevel level;
   calculateLevelForQuality(level, headerSize, maxPacketSize, qualitySetting);

   transportInfo.WantedBytesPerSecond[0].setValue(level.QualityLayer[0].BytesPerSecond);
   transportInfo.WantedPacketsPerSecond[0].setValue(level.QualityLayer[0].PacketsPerSecond);
   transportInfo.WantedMaxLossRate[0] = level.QualityLayer[0].MaxLossRate;
   if(level.QualityLayers > 1) {
      transportInfo.WantedBytesPerSecond[1].setValue(level.QualityLayer[1].BytesPerSecond);
      transportInfo.WantedPacketsPerSecond[1].setValue(level.QualityLayer[1].PacketsPerSecond);
      transportInfo.WantedMaxLossRate[1] = level.QualityLayer[1].MaxLossRate;
   }
   else {
      transportInfo.WantedBytesPerSecond[1].setValue(0);
      transportInfo.WantedPacketsPerSecond[1].setValue(0);
      transportInfo.WantedMaxLossRate[1] = 0;
   }
   if(level.QualityLayers > 2) {
      transportInfo.WantedBytesPerSecond[2].setValue(level.QualityLayer[2].BytesPerSecond);
      transportInfo.WantedPacketsPerSecond[2].setValue(level.QualityLayer[2].PacketsPerSecond);
      transportInfo.WantedMaxLossRate[2] = level.QualityLayer[2].MaxLossRate;
   }
   else {
      transportInfo.WantedBytesPerSecond[2].setValue(0);
      transportInfo.WantedPacketsPerSecond[2].setValue(0);
      transportInfo.WantedMaxLossRate[2] = 0;
   }
}


// ###### Calculate TransportInfoLevel for given quality ####################
void AdvancedAudioPacket::calculateLevelForQuality(
        TransportInfoLevel&          level,
        const cardinal               headerSize,
        const cardinal               maxPacketSize,
        const AudioQualityInterface& quality)
{
   // ====== Calculate layers and frame rate ================================
   const cardinal layers = calculateLayers(quality);
   cardinal fs1;
   if(quality.getBits() == 4)
      fs1 = calculateFrameSize(quality.getSamplingRate() / 2,AdvancedAudioFrameSize);
   else
      fs1 = calculateFrameSize(quality.getSamplingRate(),AdvancedAudioFrameSize);

   // ====== Initialize level ===============================================
   level.BytesPerSecondScale   = 65536;
   level.PacketsPerSecondScale = 65536;
   level.FramesPerSecond       = AdvancedAudioFramesPerSecond;
   level.FramesPerSecondScale  = 65536;
   level.MaxTransferDelay      = AdvancedAudioMaxTransferDelay;
   level.Quality = (card8)((quality.getBytesPerSecond() * 255) /
                              AudioQuality::HighestQuality.getBytesPerSecond());

   // ====== Initialize level's layer #0 ====================================
   level.QualityLayers = 1;
   level.QualityLayer[0].BytesPerSecond   = calculateBytesPerSecond(
                                               fs1 * AdvancedAudioFramesPerSecond,
                                               AdvancedAudioFramesPerSecond,
                                               maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize) +
                                            (AdvancedAudioMediaInfoPacketsPerSecond *
                                               (sizeof(AdvancedAudioPacket) + headerSize + sizeof(MediaInfo)));
   level.QualityLayer[0].PacketsPerSecond = calculatePacketsPerSecond(
                                               fs1 * AdvancedAudioFramesPerSecond,
                                               AdvancedAudioFramesPerSecond,
                                               maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize) +
                                            AdvancedAudioMediaInfoPacketsPerSecond;
   level.QualityLayer[0].FrameSize =
      (card32)ceil((double)level.QualityLayer[0].BytesPerSecond /
                   (double)AdvancedAudioFramesPerSecond);
   level.QualityLayer[0].MaxLossRate      = 0;
   level.QualityLayer[0].Flags            = TransportInfoLayer::TILF_Static;
   level.QualityLayer[0].pad              = 0;

   // ====== Initialize level's layer #1 ====================================
   if(layers > 1) {
      cardinal fs2 = fs1;
      if((quality.getBits() == 12) && (quality.getChannels() == 1)) {
        fs2 = calculateFrameSize(quality.getSamplingRate() / 2,AdvancedAudioFrameSize);
      }

      level.QualityLayers = 2;
      level.QualityLayer[1].BytesPerSecond   = calculateBytesPerSecond(
                                                  fs2 * AdvancedAudioFramesPerSecond,
                                                  AdvancedAudioFramesPerSecond,
                                                  maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize);
      level.QualityLayer[1].PacketsPerSecond = calculatePacketsPerSecond(
                                                  fs2 * AdvancedAudioFramesPerSecond,
                                                  AdvancedAudioFramesPerSecond,
                                                  maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize);
      level.QualityLayer[1].FrameSize =
         (card32)ceil((double)level.QualityLayer[1].BytesPerSecond /
                      (double)AdvancedAudioFramesPerSecond);
      level.QualityLayer[1].MaxLossRate      = 15;
      level.QualityLayer[1].Flags            = TransportInfoLayer::TILF_Static;
      level.QualityLayer[1].pad              = 0;
   }

   // ====== Initialize level's layer #2 ====================================
   if(layers > 2) {
      cardinal factor;
      if(quality.getBits() == 12)
         factor = 1;
      else
         factor = 2;

      level.QualityLayers = 3;
      level.QualityLayer[2].BytesPerSecond   = factor * calculateBytesPerSecond(
                                                  fs1 * AdvancedAudioFramesPerSecond,
                                                  AdvancedAudioFramesPerSecond,
                                                  maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize);
      level.QualityLayer[2].PacketsPerSecond = factor * calculatePacketsPerSecond(
                                                  fs1 * AdvancedAudioFramesPerSecond,
                                                  AdvancedAudioFramesPerSecond,
                                                  maxPacketSize, sizeof(AdvancedAudioPacket) + headerSize);
      level.QualityLayer[2].FrameSize =
         (card32)ceil((double)level.QualityLayer[2].BytesPerSecond /
                      (double)AdvancedAudioFramesPerSecond);
      level.QualityLayer[2].MaxLossRate = 30;
      level.QualityLayer[2].Flags = TransportInfoLayer::TILF_Static;
      level.QualityLayer[2].pad = 0;
   }
}


// ###### Calculate encoder's quality #######################################
AudioQuality AdvancedAudioPacket::calculateQualityForLimits(
                const AudioQualityInterface& userSetting,
                const AudioQualityInterface& inputQuality,
                const card64                 totalByteRateLimit,
                const card64                 byteRateLimitL1,
                const card64                 byteRateLimitL2,
                const card64                 byteRateLimitL3,
                const cardinal               networkQualityDecrement,
                const cardinal               headerSize,
                const cardinal               maxPacketSize)
{
   // ====== Get maximum quality ============================================
   AudioQuality networkQuality = userSetting;
   networkQuality.decrease(networkQualityDecrement);
   AudioQuality newSetting = networkQuality - inputQuality;

   // cout << "=> " << newSetting << endl;

   // ====== Calculate new quality using total byte rate limit ==============
   TransportInfoLevel level;
   while(!newSetting.isLowest()) {
      calculateLevelForQuality(level,headerSize,maxPacketSize,newSetting);
      const cardinal layers = calculateLayers(newSetting);

      // ====== Calculate total bandwidth requirements ======================
      card64 requiredTotal = level.QualityLayer[0].BytesPerSecond;
      for(cardinal i = 1;i < level.QualityLayers;i++) {
         requiredTotal += level.QualityLayer[i].BytesPerSecond;
      }

      // ====== Calculate bandwidth requirements for each level =============      
      card64 required1 = level.QualityLayer[0].BytesPerSecond;
      card64 required2;
      card64 required3;
      if(layers > 1)
         required2 = level.QualityLayer[1].BytesPerSecond;
      else
         required2 = 0;
      if(layers > 2)
         required3 = level.QualityLayer[2].BytesPerSecond;
      else
         required3 = 0;

      // ====== Check, if limits are acceptable for this level ==============
      if((requiredTotal <= totalByteRateLimit) &&
         (required1 <= byteRateLimitL1) &&
         (required2 <= byteRateLimitL2) &&
         (required3 <= byteRateLimitL3)) {
            break;
      }

/*
      cout << "(" << newSetting << ")--   => ";
      cout.flush();
      printf("req=[%Ld %Ld %Ld] <-> lim=[%Ld %Ld %Ld]\n",
         required1,required2,required3,byteRateLimitL1,byteRateLimitL2,byteRateLimitL3);
*/

      newSetting--;
   }

   // cout << "<= " << newSetting << endl;

   return(newSetting);
}


// ###### Get number of layers for given quality ############################
cardinal AdvancedAudioPacket::calculateLayers(const AudioQualityInterface& quality)
{

   cardinal layers = 1;
   if(quality.getChannels() > 1)
      layers = 2;
   if(quality.getBits() >= 16)
      layers = layers << 1;
   else if(quality.getBits() >= 12)
      layers = layers + 1;
   return(layers);
}


// ###### Calculate output frame size #######################################
cardinal AdvancedAudioPacket::calculateFrameSize(const cardinal inputBytesPerSecond,
                                                 const cardinal inputFrameSize)
{
   return((inputBytesPerSecond * inputFrameSize) /
          AudioQuality::HighestQuality.getBytesPerSecond());
}


}
