// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Audio Server Implementation                                      ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "multiaudioreader.h"
#include "socket.h"
#include "simpleaudioencoder.h"
#include "advancedaudioencoder.h"
#include "audioencoderrepository.h"
#include "rtpsender.h"

#include "rtcppacket.h"
#include "rtcpreceiver.h"
#include "rtcpabstractserver.h"
#include "sourcestateinfo.h"
#include "tools.h"
#include "randomizer.h"


#include "audioclientapppacket.h"
#include "audioserver.h"


// Verbose mode: Display some information on connections.
#define VERBOSE


// Debug mode: Print debug information
// #define QOSMGR_DEBUG

// Test: Test of flow label changes
// #define FLOWLABEL_TEST

// Test: Test of traffic class changes
// #define TC_TEST



namespace Coral {


// ###### Constructor #######################################################
AudioServer::AudioServer(QoSManager*              qosManager,
                         const cardinal           maxPacketSize,
                         CongestionManagerClient* congestionManager)
{
   Randomizer random;
   OurSSRC       = random.random32();
   QoSMgr        = qosManager;
   CongestionMgr = congestionManager;
   setMaxPacketSize(maxPacketSize);
   setLossScalability(true);
}


// ###### Destructor ########################################################
AudioServer::~AudioServer()
{
   stop();
}


// ###### Out of memory warning #############################################
void AudioServer::outOfMemoryWarning()
{
   printTimeStamp();
   cerr << "*** Out of memory! ***" << endl;
}


// ###### Send update to congestion manager #################################
void AudioServer::managementUpdate(const Client* client, User* user)
{
   // ====== Create ExtendedTransportInfo for stream ========================
   ExtendedTransportInfo streamDescription;
   user->Sender.getTransportInfo(streamDescription);

   InternetAddress ourAddress;
   user->SenderSocket.getSocketAddress(ourAddress);
   streamDescription.TotalBytesPerSecondLimit = user->BandwidthLimit;

   // ====== Do congestion management =======================================
   if(CongestionMgr != NULL) {
      CongestionMgr->setTransportInfo(user->StreamIdentifier,streamDescription);

      // ====== Receive suggestion from congestion manager ==================
      CongestionMgr->getSuggestedTransportInfo(user->StreamIdentifier,streamDescription);
   }

   // ====== Do QoS management ==============================================
   if(QoSMgr != NULL) {

#ifdef QOSMGR_DEBUG
      cout << "Wanted bandwidth to QoS-Manager:" << endl;
      cout << "Identifier: " << user->StreamIdentifier << endl;
      for(cardinal i = 0;i < streamDescription.QualityLayers;i++) {
         cout << "Layer #" << i << ":   "
              << streamDescription.WantedBytesPerSecond[i].getValue()
              << endl;
      }
#endif

      QoSMgr->updateStream(user->StreamIdentifier,streamDescription);
   }
   else {
      adaptTransportInfo((int)user->StreamIdentifier,streamDescription);
   }
}


// ###### QoSManagerInterface's adaptTransportInfo() implementation #########
void AudioServer::adaptTransportInfo(card32                streamIdentifier,
                                     ExtendedTransportInfo ti)
{
   UserSetSync.synchronized();

   multimap<const cardinal,User*>::iterator found = UserSet.find(streamIdentifier);
   if(found != UserSet.end()) {
      User* user = found->second;

      // ====== Get minimum bandwidth of QoS, CongestionMgr and limit =======
      if(CongestionMgr != NULL) {
         ExtendedTransportInfo streamDescription;
         if(CongestionMgr->getSuggestedTransportInfo(user->StreamIdentifier,streamDescription)) {
            for(cardinal i = 0;i < min((cardinal)ti.QualityLayers,TransportInfoLevel::MaxQualityLayers);i++) {
               ti.WantedBytesPerSecond[i].setValue(min(ti.WantedBytesPerSecond[i].getValue(),
                                                       streamDescription.WantedBytesPerSecond[i].getValue()));
            }
         }
      }
      ti.TotalBytesPerSecondLimit = user->BandwidthLimit;

#ifdef QOSMGR_DEBUG
      if(QoSMgr != NULL) {
         cout << "Bandwidth limits from QoS-Manager:" << endl;
         cout << "Identifier: " << streamIdentifier << endl;
         for(cardinal i = 0;i < ti.QualityLayers;i++) {
            cout << "Layer #" << i << ":   "
                 << ti.WantedBytesPerSecond[i].getValue()
                 << endl;
         }
      }
#endif


      // ====== Traffic class tests =========================================
#ifdef TC_TEST
      ti.SrcDest[0].TrafficClass = 46;
      ti.SrcDest[1].TrafficClass = 26;
      ti.SrcDest[2].TrafficClass = 34;
      ti.SrcDest[3].TrafficClass = 42;
#endif


      // ====== Flow label change tests =====================================
#ifdef FLOWLABEL_TEST
      static InternetFlow flow[3];
      static card64 lastChange = getMicroTime();

      if(getMicroTime() - lastChange > 5000000) {
         lastChange = getMicroTime();
         for(cardinal i = 0;i < 3;i++) {
            if(flow[i].getFlowLabel() != 0) {
               user->SenderSocket.freeFlow(flow[i]);
            }

            flow[i] = user->SenderSocket.allocFlow(user->Client->ClientAddress);
            if(flow[i].getFlowLabel() != 0) {
               ti.SrcDest[i].FlowLabel = flow[i].getFlowLabel();
               user->Sender.setTransportInfo(ti);
               user->Sender.getTransportInfo(ti);
               printf("Layer #%d: Flow label $%05x\n",
                      i,ti.SrcDest[i].FlowLabel);
            }
         }
      }
#endif


      // ====== Set new settings ============================================
      user->Sender.setTransportInfo(ti);

      // ====== Check, if limit is too low => bandwidth limit pause on/off ==
      if(ti.getTotalWantedBytesPerSecond() >= ti.getTotalMinWantedBytesPerSecond()) {
         if(user->ManagerLimitPause == true) {
            user->ManagerLimitPause = false;
            if((user->UserLimitPause == false) && (user->ClientPause == false)) {
#ifdef VERBOSE
               printTimeStamp();
               char str[128];
               snprintf((char*)&str,sizeof(str),"$%08x resumed due to increased manager bandwidth!",user->Client->SSRC);
               cout << str << endl;
#endif
               user->Sender.setPause(false);
            }
         }
      }
      else {
         if(user->ManagerLimitPause == false) {
            user->ManagerLimitPause = true;
            if((user->UserLimitPause == false) && (user->ClientPause == false)) {
#ifdef VERBOSE
               printTimeStamp();
               char str[128];
               snprintf((char*)&str,sizeof(str),"$%08x paused due to manager bandwidth limit!",user->Client->SSRC);
               cout << str << endl;
#endif
               user->Sender.setPause(true);
            }
         }
      }
   }
   else {
      cerr << "WARNING: AudioServer::adaptTransportInfo() - Invalid stream ID!"
           << endl;
   }
   UserSetSync.unsynchronized();
}


// ###### Add a new client ##################################################
void* AudioServer::newClient(Client* client, const char* cname)
{
   // ====== Create new user ================================================
   User* user = new User;
   if(user == NULL) {
      outOfMemoryWarning();
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }
   user->Client           = client;
   user->StreamIdentifier = 0;
   client->UserData       = user;

   // ====== Create and connect sender socket ===============================
   user->SenderSocket.create(Socket::IP,Socket::UDP);
   if(!user->SenderSocket.ready()) {
      cerr << "WARNING: AudioServer::newClient() - Unable to create socket!" << endl;      
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }      
   if(user->SenderSocket.bind() == false) {
      cerr << "WARNING: AudioServer::newClient() - Unable to bind socket!" << endl;
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }
   user->Flow = user->SenderSocket.allocFlow(client->ClientAddress);
   if(user->Flow.getFlowLabel() != 0) {
      if(user->SenderSocket.connect(user->Flow,AudioServerDefaultTrafficClass) == false) {
         if(user->SenderSocket.connect(client->ClientAddress,AudioServerDefaultTrafficClass) == false) {
            cerr << "WARNING: AudioServer::newClient() - Unable to flow-connect to client!" << endl;
            deleteClient(client,DeleteReason_Error);
            return(NULL);
         }
      }
   }
   else {
      if(user->SenderSocket.connect(client->ClientAddress,AudioServerDefaultTrafficClass) == false) {
         cerr << "WARNING: AudioServer::newClient() - Unable to connect to client!" << endl;
         deleteClient(client,DeleteReason_Error);
         return(NULL);
      }   
   }

   // ====== Create repository and encoders =================================
   user->LastSequenceNumber  = 0xffff;
   user->PosChgSeqNumber     = 0xffff;
   user->BandwidthLimit      = (card64)-1;
   user->UserLimitPause      = false;
   user->ManagerLimitPause   = false;
   user->ClientPause         = false;
   user->Repository.setAutoDelete(true);
   bool e1 = user->Repository.addEncoder(new AdvancedAudioEncoder(&user->Reader));
   bool e2 = user->Repository.addEncoder(new SimpleAudioEncoder(&user->Reader));
   if((e1 == false) || (e2 == false)) {
      outOfMemoryWarning();
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }

   // ====== Initialize RTPSender ===========================================
   user->Sender.init(OurSSRC,&user->Repository,&user->SenderSocket,MaxPacketSize);

   // ====== Add stream to QoS management ===================================
   ExtendedTransportInfo streamDescription;
   user->Sender.getTransportInfo(streamDescription);
   InternetAddress ourAddress;
   user->SenderSocket.getSocketAddress(ourAddress);
   if(QoSMgr != NULL) {
      user->StreamIdentifier = QoSMgr->getID();
   }
   else {
      user->StreamIdentifier = (integer)user;
   }

   UserSetSync.synchronized();
   UserSet.insert(pair<const cardinal,User*>(user->StreamIdentifier,user));
   UserSetSync.unsynchronized();

   if(QoSMgr != NULL) {
      QoSMgr->newStream(user->StreamIdentifier,streamDescription,this);
   }

   // ====== Add stream to congestion management ============================
   if(CongestionMgr != NULL) {
      CongestionMgr->addStream(user->StreamIdentifier,streamDescription);
   }

   // ====== Start RTPSender ================================================
   if(user->Sender.start() == false) {
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }

   // ====== Print information ==============================================
#ifdef VERBOSE
   printTimeStamp();
   char str[300];
   snprintf((char*)&str,sizeof(str),"New member $%08x added.",client->SSRC);
   cout << str << endl;
   InternetAddress sourceAddress;
   user->SenderSocket.getSocketAddress(sourceAddress);
   sourceAddress.setPrintFormat(InternetAddress::PF_Address);
   cout << "   CNAME:               " << cname << endl;
   cout << "   Source Address:      " << sourceAddress << endl;
   cout << "   Destination Address: " << (InternetAddress)client->ClientAddress << endl;
   if(user->SenderSocket.getSendFlowLabel() != 0) {
      snprintf((char*)&str,sizeof(str),"$%05x",user->SenderSocket.getSendFlowLabel());
      cout << "   Flow Label:          " << str << endl;
   }
   snprintf((char*)&str,sizeof(str),"$%02x",user->SenderSocket.getSendTrafficClass());
   cout << "   Traffic Class:       " << str << endl;
   cout << "   => We have " << getMembers() + 1 << " member(s) now!" << endl;
   if(CongestionMgr != NULL) {
      cout << "   => Congestion Manager has "
           << CongestionMgr->getStreams() << " stream(s) now." << endl;
   }
#endif
   return((void*)user);
}


// ###### Delete a client ###################################################
void AudioServer::deleteClient(Client* client, const DeleteReason reason)
{
   User* user = (User*)client->UserData;

   // ====== Remove user ====================================================
   if(user != NULL) {
      if(user->StreamIdentifier != 0) {
         UserSetSync.synchronized();
         multimap<const cardinal,User*>::iterator found =
            UserSet.find(user->StreamIdentifier);
         UserSet.erase(found);
         UserSetSync.unsynchronized();

         if(QoSMgr != NULL) {
#ifdef QOSMGR_DEBUG
            cout << "Close stream " << user->StreamIdentifier << endl;
#endif
            QoSMgr->closeStream(user->StreamIdentifier);
         }
      }
      user->Sender.stop();
      if(CongestionMgr != NULL) {
         CongestionMgr->removeStream(user->StreamIdentifier);
      }
      if(user->Flow.getFlowLabel() != 0) {
         user->SenderSocket.freeFlow(user->Flow);
      }
      delete user;
   }

   client->UserData = NULL;

   // ====== Print information ==============================================   
#ifdef VERBOSE
   char str[128];
   printTimeStamp();
   snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);

   cout << str << " removed";
   switch(reason) {
      case DeleteReason_Timeout:
         cout << " due to timeout!" << endl;
       break;
      case DeleteReason_Shutdown:
         cout << " due to server shutdown!" << endl;
        break;
      case DeleteReason_Error:
         cout << " due to transmission error!" << endl;
        break;
      default:
         cout << "." << endl;
        break;
   }
   if(reason != DeleteReason_Shutdown) {
      cout << "   => We have " << getMembers() - 1 << " member(s) now!" << endl;
      if(CongestionMgr != NULL) {
         cout << "   => Congestion Manager has "
              << CongestionMgr->getStreams() - 1 << " stream(s) now." << endl;
      }
   }
#endif
}


// ###### Handle user command ###############################################
void AudioServer::userCommand(const Client*               client,
                              User*                       user,
                              const AudioClientAppPacket* app)
{
   // ====== Check, if command has already been executed ====================
   const integer diff = (integer)app->SequenceNumber -
                           (integer)user->LastSequenceNumber;
   if(!((diff > 0) || (diff < -30000))) {
      return;
   }

   user->LastSequenceNumber = app->SequenceNumber;
   user->Sender.synchronized();

   // ====== Set quality, position, media name and bandwidth limit ==========
   if(user->Repository.selectEncoderForTypeID(app->Encoding) == true) {
      user->Repository.setSamplingRate(app->SamplingRate);
      user->Repository.setBits(app->Bits);
      user->Repository.setChannels(app->Channels);

      // ====== Select new media and set frame number =======================
      if(user->Reader.getErrorCode() == ME_NoMedia) {
#ifdef VERBOSE
            char str[128];
            printTimeStamp();
            snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);
            cout << str << " loading media <" << (const char*)&app->MediaName
                 << ">." << endl;
#endif
         if(user->Reader.openMedia((char*)&app->MediaName)) {
            user->Reader.setPosition(app->RestartPosition);
            user->PosChgSeqNumber = app->PosChgSeqNumber;
            user->MediaName       = String((const char*)&app->MediaName);
            user->Sender.leaveCorrectionLoop();
         }
      }

      else if((user->MediaName != "") && (user->MediaName != (char*)&app->MediaName)) {
#ifdef VERBOSE
         char str[128];
         printTimeStamp();
         snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);
         cout << str << " changing media to <" << (const char*)&app->MediaName
              << ">." << endl;
#endif
         user->Reader.closeMedia();
         if(user->Reader.openMedia((char*)&app->MediaName)) {
            user->Reader.setPosition(app->RestartPosition);
            user->MediaName = String((const char*)&app->MediaName);
            user->Sender.leaveCorrectionLoop();
         }
      }

      // ====== Change position, if it has been updated =====================
      const integer diff = (integer)app->PosChgSeqNumber -
                              (integer)user->PosChgSeqNumber;
      if((diff > 0) || (diff < -30000)) {
         user->PosChgSeqNumber = app->PosChgSeqNumber;
         if(app->StartPosition != (card64)-1) {
            user->Reader.setPosition(app->StartPosition);
            user->Sender.leaveCorrectionLoop();
         }
      }

      // ====== Update client's bandwidth limit =============================
      user->BandwidthLimit = (card64)app->BandwidthLimit;
   }
   else {
      char str[32];
      snprintf((char*)&str,sizeof(str),"$%04x",app->Encoding);
      cerr << "WARNING: AudioServer::userCommand() - Unsupported Encoding #"
           << str << " requested!" << endl;
   }


   // ====== Execute command ================================================
   switch(app->Status) {
      case AudioClientAppPacket::ACAS_Pause:
        user->ClientPause = true;
        user->Sender.setPause(true);
       break;
      case AudioClientAppPacket::ACAS_Play:
         {
            ExtendedTransportInfo ti;
            user->Sender.getTransportInfo(ti,false);

            if(user->BandwidthLimit >= ti.getTotalMinWantedBytesPerSecond()) {
               if(user->UserLimitPause == true) {
                  user->UserLimitPause = false;
                  if(user->ManagerLimitPause == false) {
#ifdef VERBOSE
                     printTimeStamp();
                     char str[128];
                     snprintf((char*)&str,sizeof(str),"$%08x resumed due to increased user bandwidth!",client->SSRC);
                     cout << str << endl;
#endif
                     user->Sender.setPause(false);
                  }
               }
            }
            else {
               if(user->UserLimitPause == false) {
                  user->UserLimitPause = true;
                  if(user->ManagerLimitPause == false) {
#ifdef VERBOSE
                     printTimeStamp();
                     char str[128];
                     snprintf((char*)&str,sizeof(str),"$%08x paused due to user bandwidth limit!",client->SSRC);
                     cout << str << endl;
#endif
                     user->Sender.setPause(true);
                  }
               }
            }
            if((user->ClientPause == true) && (user->UserLimitPause == false) && (user->ManagerLimitPause == false)) {
               user->ClientPause = false;
               user->Sender.setPause(false);
            }
         }
        break;
   }

   user->Sender.unsynchronized();
}


// ###### Handle app message ################################################
void AudioServer::appMessage(const Client*  client,
                             const char*    name,
                             const void*    data,
                             const cardinal dataLength)
{
   User* user = (User*)client->UserData;
   AudioClientAppPacket* app = (AudioClientAppPacket*)data;
   app->translate();
   if(app->FormatID == AudioClientAppPacket::AudioClientFormatID) {
      userCommand(client,user,app);
      managementUpdate(client,user);
   }
}


// ###### Handle SDES message ###############################################
void AudioServer::sdesMessage(const Client*  client,
                              const card8    type,
                              const char*    data,
                              const cardinal length)
{
   User* user = (User*)client->UserData;   
   if(type == RTCP_SDES_PRIV) {
      AudioClientAppPacket* app = (AudioClientAppPacket*)data;
      app->translate();
      if(app->FormatID == AudioClientAppPacket::AudioClientFormatID) {
         userCommand(client,user,app);
         managementUpdate(client,user);
      }
   }
/*
   else {
      cerr << "NOTE: AudioServer::sdesMessage() - Unsupported SDES type: "
           << (cardinal)type << endl;
   }
*/
}


// ###### Check, if client is okay ##########################################
bool AudioServer::checkClient(const Client* client)
{
   User* user = (User*)client->UserData;

   return(user->Sender.transmissionErrorDetected() == false);
}


// ###### Handle receiver report ############################################
void AudioServer::receiverReport(const Client*                   client,
                                 const RTCPReceptionReportBlock* report,
                                 const cardinal                  layer)
{
   User* user = (User*)client->UserData;

   if(LossScalability == true) {
      if(QoSMgr != NULL) {
#ifdef QOSMGR_DEBUG
         cout << "Loss rate in layer #" << layer << ": " << report->getFractionLost() << endl;
#endif
         // QoSMgr->setFractionLost(user->StreamIdentifier,report->getFractionLost(),layer);
         QoSMgr->setReport(user->StreamIdentifier,report,layer);
      }
      else {
         user->Sender.adaptQuality(report->getFractionLost(),layer);
         managementUpdate(client,user);
      }
   }
}


}
