// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Simple Audio Encoder Implementation                              ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "simpleaudioencoder.h"
#include "simpleaudiopacket.h"
#include "tools.h"
#include "audioconverter.h"


namespace Coral {


// ###### Constructor #######################################################
SimpleAudioEncoder::SimpleAudioEncoder(AudioReaderInterface* audioReader)
{
   Source           = audioReader;
   FrameBufferPos   = 0;
   FrameBufferSize  = 0;
   SendError        = 0;
   MediaInfoCounter = 0;

   setSamplingRate(AudioQuality::HighestSamplingRate);
   setChannels(2);
   setBits(16);

   NetworkQualityDecrement = 0;
   ByteRateLimit           = (card64)-1;

   FrameQualitySetting = AudioQuality(0,0,0);
   FrameBuffer = new card8[SimpleAudioPacket::SimpleAudioFrameSize];
   if(FrameBuffer == NULL)
      ErrorCode = ME_OutOfMemory;
   else
      ErrorCode = 0;
}


// ###### Destructor ########################################################
SimpleAudioEncoder::~SimpleAudioEncoder()
{
   if(FrameBuffer != NULL) {
      delete FrameBuffer;
      FrameBuffer = NULL;
   }
}


// ###### Activate encoder ##################################################
void SimpleAudioEncoder::activate()
{
   reset();
}


// ###### Deactivate encoder ################################################
void SimpleAudioEncoder::deactivate()
{
   reset();
}


// ###### Reset encoder #####################################################
void SimpleAudioEncoder::reset()
{
   NetworkQualityDecrement = 0;
   ByteRateLimit           = (card64)-1;

   FrameQualitySetting = AudioQuality(0,0,0);
   FrameBufferPos   = 0;
   FrameBufferSize  = 0;
   SendError        = 0;
   MediaInfoCounter = 0;
   if(FrameBuffer != NULL)
      ErrorCode = 0;
   else
      ErrorCode = ME_OutOfMemory;
}


// ###### Get encoder's type ID #############################################
const card16 SimpleAudioEncoder::getTypeID() const
{
   return(SimpleAudioPacket::SimpleAudioTypeID);
}


// ###### Get encoder's type name ###########################################
const char* SimpleAudioEncoder::getTypeName() const
{
   return((const char*)&SimpleAudioPacket::SimpleAudioTypeName);
}


// ###### Get TransportInfo #################################################
void SimpleAudioEncoder::getTransportInfo(TransportInfo& transportInfo,
                                          const cardinal headerSize,
                                          const cardinal maxPacketSize,
                                          const bool     calculateLevels) const
{
   SimpleAudioPacket::getTransportInfo(transportInfo,headerSize,maxPacketSize,
                                       (AudioQuality)*this,
                                       FrameQualitySetting,
                                       calculateLevels);
   transportInfo.TotalBytesPerSecondLimit   = (card64)-1;
   transportInfo.TotalPacketsPerSecondLimit = (card32)-1;
   transportInfo.TotalFramesPerSecondLimit  = (card32)-1;
}


// ###### Set TransportInfo #################################################
void SimpleAudioEncoder::setTransportInfo(TransportInfo& transportInfo,
                                          const cardinal headerSize,
                                          const cardinal maxPacketSize,
                                          const bool     calculateLevels)
{
   // ====== Get constants ==================================================
   SimpleAudioPacket::getTransportInfoConstants(transportInfo,
                                                headerSize,maxPacketSize,
                                                FrameQualitySetting,
                                                calculateLevels);

   // ====== Set byte rate limit ============================================
   ByteRateLimit = min(transportInfo.WantedBytesPerSecond[0].getValue(),
                       transportInfo.TotalBytesPerSecondLimit);

   // ====== Return current settings ========================================
   const AudioQuality newSetting = SimpleAudioPacket::calculateQualityForLimits(
      (AudioQuality)*this,
      (AudioQuality)*Source,
      ByteRateLimit, NetworkQualityDecrement, headerSize, maxPacketSize);

   SimpleAudioPacket::calculateLevelForQuality(
      transportInfo.CurrentSetting, 
      headerSize, maxPacketSize, newSetting);
   transportInfo.CurrentSetting.LevelUp   = SimpleAudioPacket::SimpleAudioQualityLevels - 1;
   transportInfo.CurrentSetting.LevelDown = 0;

}


// ###### Adapt quality #####################################################
void SimpleAudioEncoder::adaptQuality(const double   fractionLost,
                                      const cardinal layer)
{
   if(fractionLost > SimpleAudioPacket::SimpleAudioUpperLossThreshold) {
      if(NetworkQualityDecrement < SimpleAudioPacket::SimpleAudioQualityLevels - 1) {
         NetworkQualityDecrement++;
      }
   }
   else if(fractionLost < SimpleAudioPacket::SimpleAudioLowerLossThreshold) {
      if(NetworkQualityDecrement > 0) {
         NetworkQualityDecrement--;
      }
   }
}


// ###### Prepare frame for packaging #######################################
bool SimpleAudioEncoder::prepareNextFrame(const cardinal headerSize,
                                          const cardinal maxPacketSize,
                                          const cardinal flags)
{
   if(ErrorCode == ME_OutOfMemory)
      return(false);

   // ====== Calculate frame parameters =====================================
   FrameBufferPos      = 0;
   FrameBufferSize     = 0;
   FramePosition       = Source->getPosition();
   FrameMaxPosition    = Source->getMaxPosition();
   const AudioQuality inputQuality(*Source);
   FrameQualitySetting = SimpleAudioPacket::calculateQualityForLimits(
                            (AudioQuality)*this,(AudioQuality)*Source,
                            ByteRateLimit,NetworkQualityDecrement,
                            headerSize,maxPacketSize);
   FrameQualitySetting.setByteOrder(BIG_ENDIAN);

   // ====== Read frame from AudioReader ====================================
   cardinal len = SimpleAudioPacket::calculateFrameSize(
                      inputQuality.getBytesPerSecond(),
                      SimpleAudioPacket::SimpleAudioFrameSize);
   if((Source->getPosition() < Source->getMaxPosition()) && (Source->getNextBlock((void*)FrameBuffer,len) == len)) {
      // Check, if conversion is necessary
      if(inputQuality != FrameQualitySetting) {
         AudioConverter(inputQuality,FrameQualitySetting,
                        FrameBuffer,FrameBuffer,len,len);
      }
      FrameBufferSize = getAlignedLength(inputQuality,FrameQualitySetting,len);

      ErrorCode = ME_NoError;
      SendError = 0;
      MediaInfoCounter--;
      return(true);
   }
   else {
      // Source has an error -> send error message
      card8 error = (card8)Source->getErrorCode();
      if(error == 0) {
         if(Source->getPosition() >= Source->getMaxPosition()) {
            error = ME_EOF;
         }
      }
      if(error > 0) {
         if(SendError == 0) {
            ErrorCode       = error;
            FrameBufferPos  = 0;
            FrameBufferSize = 4;
            FrameBuffer[0]  = 0x00;
            SendError       = SimpleAudioPacket::SimpleAudioFramesPerSecond;
            return(true);
         }
         if(SendError > 0) SendError--;
         return(false);
      }
   }
   return(false);
}


// ###### Get next packet from current frame ################################
cardinal SimpleAudioEncoder::getNextPacket(EncoderPacket* encoderPacket)
{
   // ====== Create packet ===============================================
   SimpleAudioPacket* packet  = (SimpleAudioPacket*)encoderPacket->Buffer;
   card8*             data    = (card8*)FrameBuffer;
   encoderPacket->Layer       = 0;
   encoderPacket->PayloadType = (card8)(SimpleAudioPacket::SimpleAudioTypeID & 0x00ff);

   packet->FormatID     = SimpleAudioPacket::SimpleAudioFormatID;
   packet->Position     = FramePosition;
   packet->MaxPosition  = FrameMaxPosition;

   packet->SamplingRate = FrameQualitySetting.getSamplingRate();
   packet->Bits         = FrameQualitySetting.getBits();
   packet->Channels     = FrameQualitySetting.getChannels();
   packet->ErrorCode    = ErrorCode;
   packet->Flags        = SimpleAudioPacket::SAF_Data;

   // ====== Get payload data ===============================================
   if(FrameBufferPos < FrameBufferSize) {

      // ====== Ensure correct alignment of data ============================
      cardinal bytes = min(encoderPacket->MaxLength - sizeof(SimpleAudioPacket),
                           FrameBufferSize - FrameBufferPos);
      if(packet->Bits == 12) {
         if(packet->Channels == 2)
            bytes = bytes - (bytes % 6);
         else
            bytes = bytes - (bytes % 3);
      }
      if(bytes == 0) {
         return(0);
      }

      for(cardinal i = 0;i < bytes;i++) {
         packet->Data[i] = data[FrameBufferPos++];
      }

      // ====== Translate byte order ========================================
      packet->translate();
      return(bytes + sizeof(SimpleAudioPacket));
   }
   else if(MediaInfoCounter <= 0) {
      MediaInfoCounter = 2 + (SimpleAudioPacket::SimpleAudioFramesPerSecond /
                                 SimpleAudioPacket::SimpleAudioMediaInfoPacketsPerSecond);

      if(encoderPacket->MaxLength >= sizeof(SimpleAudioPacket) + sizeof(MediaInfo)) {
         MediaInfo* media = (MediaInfo*)&packet->Data[0];
         Source->getMediaInfo(*media);
         packet->Flags = SimpleAudioPacket::SAF_MediaInfo;
         packet->translate();
         return(sizeof(MediaInfo) + sizeof(SimpleAudioPacket));
      }
      else {
         cerr << "WARNING: SimpleAudioEncoder::getNextPacket() - "
              << "Packet size too low for media info!" << endl;
      }
   }
   return(0);
}


}
