// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Congestion Monitor Implementation                                ####
// ####                                                                  ####
// #### Version 1.00  --  February 16, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "networkmonitorinterface.h"
#include "networkmonitor.h"
#include "congestionmonitorreport.h"
#include "timedthread.h"
#include "socket.h"
#include "tools.h"
#include "randomizer.h"
#include "breakdetector.h"
#include "congestionmonitor.h"


// Verbose mode: Display some information on connections.
#define VERBOSE


namespace Coral {


// ###### Constructor #######################################################
CongestionMonitor::CongestionMonitor(NetworkMonitorInterface* monitor,
                                     const card64             bandwidth,
                                     const card8              flags)
   : TimedThread(250000)
{
   setTimerCorrection(0);
   Monitor   = monitor;
   Bandwidth = bandwidth;
   Flags     = flags;

   Randomizer random;
   SequenceNumber = random.random32();
}


// ###### Destructor ########################################################
CongestionMonitor::~CongestionMonitor()
{
   stop();
}


// ###### TimedThread's timerEvent() implementation #########################
void CongestionMonitor::timerEvent()
{
   synchronized();

   // ====== Get report =====================================================
   CongestionMonitorReport report;
   Monitor->moveReport(&report);
   report.Bandwidth  = Bandwidth;
   report.SeqNum     = SequenceNumber++;
   report.Congestion = 
      (card32)((((double)report.BytesRaw * 1000000.0) /
                 (double)report.Interval) * 65536.0 / (double)Bandwidth);
   report.CongestionMonitorFlags = Flags;

/*
   printf("%Ld of %Ld => %f%%\n",report.BytesRaw,
                                 (card64)((double)Bandwidth * ((double)report.Interval / 1000000.0)),
                                 report.Congestion * 100.0 / 65536.0);
*/

   report.translate();

   // ====== Send report to all receivers ===================================
   multiset<Socket*>::iterator setIterator = ReportReceiverSet.begin();
   while(setIterator != ReportReceiverSet.end()) {
      Socket* receiver = *setIterator;
      const ssize_t result = receiver->write(&report,sizeof(report));
      if(result != sizeof(report)) {
         removeReportReceiver(receiver);
      }
      setIterator++;
   }

   unsynchronized();
}


// ###### Add receiver to monitor ###########################################
bool CongestionMonitor::addReportReceiver(Socket* receiver)
{
   bool added = false;
   synchronized();
   multiset<Socket*>::iterator found = ReportReceiverSet.find(receiver);
   if(found == ReportReceiverSet.end()) {
      ReportReceiverSet.insert(receiver);
      added = true;
   }
   unsynchronized();

#ifdef VERBOSE
   if(added) {
      InternetAddress address;
      receiver->getPeerAddress(address);
      cout << "Receiver #" << getReceivers() << " from " << address
           << " added." << endl;
   }
#endif
   return(added);
}


// ###### Remove receiver from monitor ######################################
void CongestionMonitor::removeReportReceiver(Socket* receiver)
{
#ifdef VERBOSE
   cout << "Receiver #" << getReceivers() << " removed." << endl;
#endif          

   synchronized();
   multiset<Socket*>::iterator found = ReportReceiverSet.find(receiver);
   if(found != ReportReceiverSet.end()) {
      ReportReceiverSet.erase(found);
      delete receiver;
   }
   unsynchronized();
}


}
