// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Network Analyzer                                                 ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "netlogreader.h"
#include "strings.h"


#include <fstream.h>


using namespace Coral;


cardinal         PageNumber = 1;
cardinal         MaxStreams = 1000;
StreamReport**   Stream;
String**         StreamName;
cardinal         StreamCount = 0;
cardinal         Usage       = 0;


// ###### Write a GNUplot page initialization header ########################
void newpage(ostream&    os,
             const char* title,
             const char* xlabel,
             const char* ylabel,
             const char* info,
             const char* subtitle)
{
   cout << "Page #" << PageNumber++ << ": " << title << " - " << subtitle << endl;
   os << "set timestamp \"" << info << "\" 0,0 \"Times-Roman,12\"" << endl;
   os << "set title \"" << title << "\" \"Times-Roman,14\"" << endl;
   os << "set grid" << endl;
   os << "set ylabel \"" << ylabel << "\"" << endl;
   os << "set xlabel \"" << xlabel << "\"" << endl;
}


// ###### Write a GNUplot plot command line #################################
void plot(const char*    pagetitle,
          const char*    xlabel,
          const char*    ylabel,
          const bool     ownPage,
          const char*    info,
          ostream&       os,
          bool&          firstUsage,
          const char*    name,
          const cardinal index,
          const char*    title,
          const card64   xRangeMin = 0,
          const card64   xRangeMax = 0,
          const card64   yRangeMin = 0,
          const card64   yRangeMax = 0)
{
   if((firstUsage == true) || (ownPage == true)) {
      newpage(os,pagetitle,xlabel,ylabel,info,title);
      firstUsage = false;
      os << "plot ";
      if(xRangeMin != xRangeMax) {
         os << "[" << xRangeMin << ":" << xRangeMax << "] ";
      }
      else {
         os << "[] ";
      }
      if(yRangeMin != yRangeMax) {
         const double yRangeMax2 = 1.1 * (double)yRangeMax;
         os << "[" << yRangeMin << ":" << yRangeMax2 << "] ";
      }
      else {
         os << "[] ";
      }
      Usage = 0;
   }
   else {
      os << ", ";
   }
   Usage++;
   os << "\"" << name <<  "\" using 1:" << index << " title \"" << title << "\" with lines ls " << Usage;
   if(ownPage) {
      os << endl;
   }
}


// ###### Division with check for division by zero ##########################
card64 safeDivision(const card64 p, const card64 q)
{
   if(q == 0)
      return(0);
   return(p / q);
}


// ###### Convert value from input interval to output interval ##############
double conv(const card64 value,
            const card64 inputInterval,
            const card64 outputInterval)
{
   const double newValue = ((double)value * (double)(outputInterval / 1000)) /
                              (double)(inputInterval / 1000);
   return(newValue);
}


// ###### Add stream to set #################################################
void addStream(const StreamReport& report)
{
   for(cardinal i = 0;i < StreamCount;i++) {
      if(*(Stream[i]) == report) {
         *(Stream[i]) = report;
         return;
      }
   }
   if(StreamCount >= MaxStreams) {
      cerr << "WARNING: addStream() - MaxStreams setting too low!" << endl;
   }
   else {
      *(Stream[StreamCount]) = report;
      StreamCount++;
   }
}


// ###### Update stream in set ##############################################
void updateStream(const StreamReport& report,
                  const card64        time,
                  const card64        inputInterval,
                  const card64        outputInterval)
{
   for(cardinal j = 0;j < StreamCount;j++) {
      if(*(Stream[j]) == report) {
         for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
            Stream[j]->BytesRaw[i]     += report.BytesRaw[i];
            Stream[j]->BytesPayload[i] += report.BytesPayload[i];
            Stream[j]->Packets[i]      += report.Packets[i];
         }

         ofstream os(StreamName[j]->getData(),ios::app);
         if(os.good()) {
            os << time / 1000000.0 << " ";
            card64   bytesRaw     = 0;
            card64   bytesPayload = 0;
            cardinal packets      = 0;

            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               os << conv(report.BytesRaw[i],inputInterval,outputInterval) << " ";
               bytesRaw += report.BytesRaw[i];
            }
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               os << conv(report.BytesPayload[i],inputInterval,outputInterval) << " ";
               bytesPayload += report.BytesPayload[i];
            }

            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               os << conv(report.Packets[i],inputInterval,outputInterval) << " ";
               packets += report.Packets[i];
            }

            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               const cardinal avg = safeDivision(report.BytesRaw[i],report.Packets[i]);
               os << avg << " ";
            }

            os << conv(bytesRaw,inputInterval,outputInterval) << " "
               << conv(bytesPayload,inputInterval,outputInterval) << " "
               << conv(packets,inputInterval,outputInterval) << " "
               << safeDivision(bytesRaw,packets) << endl;
         }
         else {
            cerr << "ERROR: updateStream() - Write error!" << endl;
            exit(1);
         }

         return;
      }
   }
   cerr << "ERROR: File structure error!" << endl;
}


// ###### Main program ######################################################
int main(int argc, char** argv)
{
   cardinal btcp      = 0;
   cardinal budp      = 0;
   cardinal btcpdata  = 0;
   cardinal budpdata  = 0;
   cardinal bicmp4    = 0;
   cardinal bicmp6    = 0;
   cardinal bipv4     = 0;
   cardinal bipv6     = 0;
   cardinal bip       = 1;
   cardinal ptcp      = 0;
   cardinal pudp      = 0;
   cardinal picmp4    = 0;
   cardinal picmp6    = 0;
   cardinal pipv4     = 0;
   cardinal pipv6     = 0;
   cardinal pip       = 0;
   cardinal ltcp      = 0;
   cardinal ludp      = 0;
   cardinal licmp4    = 0;
   cardinal licmp6    = 0;
   cardinal lipv4     = 0;
   cardinal lipv6     = 0;
   cardinal lip       = 0;
   cardinal bclasses  = 0;
   cardinal pclasses  = 0;
   cardinal lclasses  = 0;
   cardinal b1streams = 0;
   cardinal b2streams = 0;
   cardinal pstreams  = 0;
   cardinal lstreams  = 0;
   bool   ownPage        = false;
   bool   ownPageStreams = false;
   bool   autoScale      = true;
   bool   optForceIPv4   = false;
   char*  info           = NULL;
   cardinal format       = InternetAddress::PF_Full;
   card64 outputInterval = 1000000;

   cardinal bclass[TrafficClassValues::MaxValues];
   cardinal pclass[TrafficClassValues::MaxValues];
   cardinal lclass[TrafficClassValues::MaxValues];
   cardinal bclassvalid[TrafficClassValues::MaxValues];
   cardinal pclassvalid[TrafficClassValues::MaxValues];
   cardinal lclassvalid[TrafficClassValues::MaxValues];
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      bclass[i] = pclass[i] = lclass[i] = 0;
      bclassvalid[i] = pclassvalid[i] = lclassvalid[i] = false;
   }


   // ====== Get arguments ==================================================
   if(argc < 3) {
      cerr << "Usage: " << argv[0] << endl
           << "       [Input file] [Output prefix] "
              "{-interval=milliseconds} "
              "{[+/-]{|pkt|len}tcp} "
              "{[+/-]tcpdata} {[+/-]{|pkt|len}udp} {udpdata} "
              "{[+/-]{|pkt|len}icmp4} {[+/-]{|pkt|len}icmp6} {[+/-]{|pkt|len}ip} {[+/-]{|pkt|len}ipv4} "
              "{[+/-]{|pkt|len}ipv6} {[+/-]{|pkt|len}classes} {[+/-]{|pkt|len}streams} "
              "{[+/-]streamsraw} {streamspayload} "
              "{[+/-]address} {[+/-]hostname} {[+/-]ownpage} "
              "{[+/-]ownpagestreams} {[+/-]autoscale} {+all} "
              "{-info=infostring} {-force-ipv4}"
           << endl;
      exit(1);
   }

   // ====== Get arguments ==================================================
   for(cardinal i = 3;i < (cardinal)argc;i++) {
      if(!(strcasecmp(argv[i],"-tcp")))           btcp     = 0;
      else if(!(strcasecmp(argv[i],"+tcp")))      btcp     = 1;
      else if(!(strcasecmp(argv[i],"-tcpdata")))  btcpdata = 0;
      else if(!(strcasecmp(argv[i],"+tcpdata")))  btcpdata = 1;
      else if(!(strcasecmp(argv[i],"-udp")))      budp     = 0;
      else if(!(strcasecmp(argv[i],"+udp")))      budp     = 1;
      else if(!(strcasecmp(argv[i],"-udpdata")))  budpdata = 0;
      else if(!(strcasecmp(argv[i],"+udpdata")))  budpdata = 1;
      else if(!(strcasecmp(argv[i],"-icmp4")))    bicmp4   = 0;
      else if(!(strcasecmp(argv[i],"+icmp4")))    bicmp4   = 1;
      else if(!(strcasecmp(argv[i],"-icmp6")))    bicmp6   = 0;
      else if(!(strcasecmp(argv[i],"+icmp6")))    bicmp6   = 1;
      else if(!(strcasecmp(argv[i],"-ipv6")))     bipv6    = 0;
      else if(!(strcasecmp(argv[i],"+ipv6")))     bipv6    = 1;
      else if(!(strcasecmp(argv[i],"-ipv4")))     bipv4    = 0;
      else if(!(strcasecmp(argv[i],"+ipv4")))     bipv4    = 1;
      else if(!(strcasecmp(argv[i],"-ip")))       bip      = 0;
      else if(!(strcasecmp(argv[i],"+ip")))       bip      = 1;

      else if(!(strcasecmp(argv[i],"-pkttcp")))   ptcp     = 0;
      else if(!(strcasecmp(argv[i],"+pkttcp")))   ptcp     = 1;
      else if(!(strcasecmp(argv[i],"-pktudp")))   pudp     = 0;
      else if(!(strcasecmp(argv[i],"+pktudp")))   pudp     = 1;
      else if(!(strcasecmp(argv[i],"-pkticmp4"))) picmp4   = 0;
      else if(!(strcasecmp(argv[i],"+pkticmp4"))) picmp4   = 1;
      else if(!(strcasecmp(argv[i],"-pkticmp6"))) picmp6   = 0;
      else if(!(strcasecmp(argv[i],"+pkticmp6"))) picmp6   = 1;
      else if(!(strcasecmp(argv[i],"-pktipv6")))  pipv6    = 0;
      else if(!(strcasecmp(argv[i],"+pktipv6")))  pipv6    = 1;
      else if(!(strcasecmp(argv[i],"-pktipv4")))  pipv4    = 0;
      else if(!(strcasecmp(argv[i],"+pktipv4")))  pipv4    = 1;
      else if(!(strcasecmp(argv[i],"-pktip")))    pip      = 0;
      else if(!(strcasecmp(argv[i],"+pktip")))    pip      = 1;

      else if(!(strcasecmp(argv[i],"-lentcp")))   ltcp     = 0;
      else if(!(strcasecmp(argv[i],"+lentcp")))   ltcp     = 1;
      else if(!(strcasecmp(argv[i],"-lenudp")))   ludp     = 0;
      else if(!(strcasecmp(argv[i],"+lenudp")))   ludp     = 1;
      else if(!(strcasecmp(argv[i],"-lenicmp4"))) licmp4   = 0;
      else if(!(strcasecmp(argv[i],"+lenicmp4"))) licmp4   = 1;
      else if(!(strcasecmp(argv[i],"-lenicmp6"))) licmp6   = 0;
      else if(!(strcasecmp(argv[i],"+lenicmp6"))) licmp6   = 1;
      else if(!(strcasecmp(argv[i],"-lenipv6")))  lipv6    = 0;
      else if(!(strcasecmp(argv[i],"+lenipv6")))  lipv6    = 1;
      else if(!(strcasecmp(argv[i],"-lenipv4")))  lipv4    = 0;
      else if(!(strcasecmp(argv[i],"+lenipv4")))  lipv4    = 1;
      else if(!(strcasecmp(argv[i],"-lenip")))    lip      = 0;
      else if(!(strcasecmp(argv[i],"+lenip")))    lip      = 1;

      else if(!(strcasecmp(argv[i],"-classes")))        bclasses = 0;
      else if(!(strcasecmp(argv[i],"+classes")))        bclasses = 1;
      else if(!(strcasecmp(argv[i],"-pktclasses")))     pclasses = 0;
      else if(!(strcasecmp(argv[i],"+pktclasses")))     pclasses = 1;
      else if(!(strcasecmp(argv[i],"-lenclasses")))     lclasses = 0;
      else if(!(strcasecmp(argv[i],"+lenclasses")))     lclasses = 1;

      else if(!(strcasecmp(argv[i],"-streams")))        { b1streams = 0; b2streams = 0; }
      else if(!(strcasecmp(argv[i],"-streamsraw")))     b1streams = 0;
      else if(!(strcasecmp(argv[i],"-streamspayload"))) b2streams = 0;
      else if(!(strcasecmp(argv[i],"+streams")))        { b1streams = 1; b2streams = 1; }
      else if(!(strcasecmp(argv[i],"+streamsraw")))     b1streams = 1;
      else if(!(strcasecmp(argv[i],"+streamspayload"))) b2streams = 1;
      else if(!(strcasecmp(argv[i],"-pktstreams")))     pstreams  = 0;
      else if(!(strcasecmp(argv[i],"+pktstreams")))     pstreams  = 1;
      else if(!(strcasecmp(argv[i],"-lenstreams")))     lstreams  = 0;
      else if(!(strcasecmp(argv[i],"+lenstreams")))     lstreams  = 1;

      else if(!(strcasecmp(argv[i],"-address")))        format &= ~InternetAddress::PF_Address;
      else if(!(strcasecmp(argv[i],"+address")))        format |= InternetAddress::PF_Address;
      else if(!(strcasecmp(argv[i],"-hostname")))       format &= ~InternetAddress::PF_Hostname;
      else if(!(strcasecmp(argv[i],"+hostname")))       format |= InternetAddress::PF_Hostname;

      else if(!(strcasecmp(argv[i],"-ownpage")))        ownPage        = false;
      else if(!(strcasecmp(argv[i],"+ownpage")))        ownPage        = true;
      else if(!(strcasecmp(argv[i],"-autoscale")))      autoScale      = false;
      else if(!(strcasecmp(argv[i],"+autoscale")))      autoScale      = true;
      else if(!(strcasecmp(argv[i],"-ownpagestreams"))) ownPageStreams = false;
      else if(!(strcasecmp(argv[i],"+ownpagestreams"))) ownPageStreams = true;

      else if(!(strncasecmp(argv[i],"-maxstreams=",12))) MaxStreams    = atol(&argv[i][12]);
      else if(!(strcasecmp(argv[i],"-force-ipv4")))      optForceIPv4  = true;
      else if(!(strncasecmp(argv[i],"-info=",6)))        info          = &argv[i][6];
      else if(!(strncasecmp(argv[i],"-interval=",10))) {
         outputInterval = (card64)1000 * (card64)atol(&argv[i][10]);
      }

      else if(!(strcasecmp(argv[i],"+all"))) {
         btcp = btcpdata = budp = budpdata = bicmp4 = bicmp6 = bipv6 = bipv4 = bip = 1;
         ptcp = pudp = picmp4 = picmp6 = pipv6 = pipv4 = pip = 1;
         ltcp = ludp = licmp4 = licmp6 = lipv6 = lipv4 = lip = 1;
         bclasses = pclasses = lclasses = 1;
         b1streams = b2streams = lstreams = pstreams = 1;
      }
      else {
         cerr << "Wrong parameter: " << argv[i] << endl;
         exit(1);
      }
   }
   if( (!(format & InternetAddress::PF_Address)) && (!(format & InternetAddress::PF_Hostname)) ) {
      format = InternetAddress::PF_Address;
   }
   if(optForceIPv4) {
      if(InternetAddress::UseIPv6 == true) {
         InternetAddress::UseIPv6 = false;
         cerr << "NOTE: IPv6 support disabled!" << endl;
      }
   }


   // ====== Allocate stream buffer =========================================
   if(MaxStreams < 100) {
      MaxStreams = 100;
   }
   if(outputInterval < 10000) {
      outputInterval = 10000;
   }
   Stream    = new StreamReport*[MaxStreams];
   StreamName = new String*[MaxStreams];
   if((Stream == NULL) || (StreamName == NULL)) {
      cerr << "ERROR: Out of memory!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < MaxStreams;i++) {
      Stream[i] = new StreamReport;
      StreamName[i] = new String;
      if((Stream[i] == NULL) || (StreamName[i] == NULL)) {
         cerr << "ERROR: Out of memory!" << endl;
         exit(1);
      }
   }

   // ====== Open files =====================================================
   NetLogReader* reader = new NetLogReader(argv[1]);
   if(reader->ready() == false) {
      cerr << "ERROR: Unable to read input!" << endl;
      exit(1);
   }
   String dname = argv[2] + (String)".data";
   ofstream dataos(dname.getData());
   if(dataos.good() == false) {
      cerr << "ERROR: Unable to write output!" << endl;
      exit(1);
   }
   String gname = argv[2] + (String)".gp";
   ofstream gpos(gname.getData());
   if(gpos.good() == false) {
      cerr << "ERROR: Unable to write output!" << endl;
      exit(1);  
   }
   const card64 inputInterval = reader->getInterval();

   // ====== Get stream overview ============================================
   cout << "Network Analyzer - Copyright (C) 1999-2001 Thomas Dreibholz" << endl;
   cout << "-----------------------------------------------------------" << endl;
   cout << endl;
   cout << "Input file:         " << argv[1] << endl;
   cout << "Input interval:     " << inputInterval  << " [s]" << endl;
   cout << "Output interval:    " << outputInterval << " [s]" << endl;
   cout << "Script file:        " << gname   << endl;
   cout << "Main data file:     " << dname   << endl;
   cout << "Stream data files:  " << argv[2] << "-<src>-<dest>.data" << endl;
   cout << endl;
   cout << "Analyzing streams..." << endl;
   card64   tmin = 0;
   card64   tmax = 0;
   card64   brmax[TrafficClassValues::MaxValues];
   card64   bpmax[TrafficClassValues::MaxValues];
   cardinal pmax[TrafficClassValues::MaxValues];
   cardinal almax[TrafficClassValues::MaxValues];
   card64   tbrmax = 0;
   card64   tbpmax = 0;
   cardinal tpmax  = 0;
   cardinal tamax  = 0;
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      brmax[i] = 0;
      bpmax[i] = 0;
      pmax[i]  = 0;
      almax[i] = 0;
   }
   while(!reader->eof()) {
      NetworkMonitorReport report;
      if(reader->readNextReport(&report)) {
         StreamReport sr;
         while(reader->readNextStreamReport(&sr)) {
            addStream(sr);
            card64   br = 0;
            card64   bp = 0;
            cardinal p  = 0;
            cardinal a  = 0;
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               if(sr.BytesRaw[i] >     brmax[i]) brmax[i] = sr.BytesRaw[i];
               if(sr.BytesPayload[i] > bpmax[i]) bpmax[i] = sr.BytesPayload[i];
               if(sr.Packets[i] >      pmax[i])  pmax[i] = sr.Packets[i];
               const cardinal al = safeDivision(sr.BytesRaw[i],sr.Packets[i]);
               if(al > almax[i])  almax[i] = al;
               br += sr.BytesRaw[i];
               bp += sr.BytesPayload[i];
               p  += sr.Packets[i];
               a  += al;
            }
            if(br > tbrmax) tbrmax = br;
            if(bp > tbpmax) tbpmax = bp;
            if(p > tpmax)   tpmax  = p;
            if(a > tamax)   tamax  = a;
         }

         if(tmin == 0) {
            tmin = report.TimeStamp;
         }

         // If there is no time stamp (==0), calculate tmax from old tmax + Interval.
         if(report.TimeStamp == 0) {
            tmax += report.Interval;
         }
         else {
            tmax = report.TimeStamp - tmin;
         }
      }
   }
   const cardinal maxTime = (cardinal)(tmax / 1000000);

   // ====== Generate stream names ==========================================
   for(cardinal i = 0;i < StreamCount;i++) {
      InternetAddress source      = InternetAddress(Stream[i]->Source);
      InternetAddress destination = InternetAddress(Stream[i]->Destination);
      source.setPrintFormat((InternetAddress::PrintFormat)format);
      destination.setPrintFormat((InternetAddress::PrintFormat)format); 
      char str[32];
      if(Stream[i]->FlowLabel != 0) {
         snprintf((char*)&str,sizeof(str),"[$%05x]",Stream[i]->FlowLabel);
      }
      else {
         str[0] = 0x00;
      }
      *(StreamName[i]) = (String)argv[2] +
                            "-" + source.getAddressString() +
                            str +
                            "-" + destination.getAddressString() +
                            ".data";
      remove(StreamName[i]->getData());
   }
   cout << endl;
   cout << "Monitoring time:    " << maxTime << " [s]" << endl;
   cout << "Number of streams:  " << StreamCount << endl;
   cout << endl;


   // ====== AutoScale: Let GNUplot choose best scale values ================
   if(autoScale == true) {
      tbrmax = 0;
      tbpmax = 0;
      tpmax  = 0;
      tamax  = 0;
      for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
         brmax[i] = 0;
         bpmax[i] = 0;
         pmax[i]  = 0;
         almax[i] = 0;
      }
   }


   // ====== Generate data output ===========================================
   cout << "Generating output..." << endl << endl;
   card64 time = 0;
   reader->reset();
   while(!reader->eof()) {
      NetworkMonitorReport report;
      if(reader->readNextReport(&report)) {
         dataos << time / 1000000.0 << " ";

         // ====== Write data ===============================================
         cardinal pos = 2;
         if(bipv4) {
            dataos << conv(report.BytesIPv4,report.Interval,outputInterval) << " ";
            bipv4 = pos++;
         }
         if(bipv6) {
            dataos << conv(report.BytesIPv6,report.Interval,outputInterval) << " ";
            bipv6 = pos++;
         }
         if(bip) {
            dataos << conv(report.BytesIPv4 + report.BytesIPv6,report.Interval,outputInterval) << " ";
            bip = pos++;
         }
         if(budp) {
            dataos << conv(report.BytesUDP,report.Interval,outputInterval) << " ";
            budp = pos++;
         }
         if(budpdata) {
            dataos << conv(report.BytesUDPData,report.Interval,outputInterval) << " ";
            budpdata = pos++;
         }
         if(btcp) {
            dataos << conv(report.BytesTCP,report.Interval,outputInterval) << " ";
            btcp = pos++;
         }
         if(btcpdata) {
            dataos << conv(report.BytesTCPData,report.Interval,outputInterval) << " ";
            btcpdata = pos++;
         }
         if(bicmp4) {
            dataos << conv(report.BytesICMP4,report.Interval,outputInterval) << " ";
            bicmp4 = pos++;
         }
         if(bicmp6) {
            dataos << conv(report.BytesICMP6,report.Interval,outputInterval) << " ";
            bicmp6 = pos++;
         }
         if(bclasses) {
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               dataos << conv(report.ClassBytes[i],report.Interval,outputInterval) << " ";
               if(report.ClassBytes[i] != 0) {
                  bclassvalid[i] = true;
               }
               bclass[i] = pos++;
            }
         }

         if(pipv4) {
            dataos << conv(report.PacketsIPv4,report.Interval,outputInterval) << " ";
            pipv4 = pos++;
         }
         if(pipv6) {
            dataos << conv(report.PacketsIPv6,report.Interval,outputInterval) << " ";
            pipv6 = pos++;
         }
         if(pip) {
            dataos << conv(report.PacketsIPv4 + report.PacketsIPv6,report.Interval,outputInterval) << " ";
            pip = pos++;
         }
         if(pudp) {
            dataos << conv(report.PacketsUDP,report.Interval,outputInterval) << " ";
            pudp = pos++;
         }
         if(ptcp) {
            dataos << conv(report.PacketsTCP,report.Interval,outputInterval) << " ";
            ptcp = pos++;
         }
         if(picmp4) {
            dataos << conv(report.PacketsICMP4,report.Interval,outputInterval) << " ";
            picmp4 = pos++;
         }
         if(picmp6) {
            dataos << conv(report.PacketsICMP6,report.Interval,outputInterval) << " ";
            picmp6 = pos++;
         }
         if(pclasses) {
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               dataos << conv(report.ClassPackets[i],report.Interval,outputInterval) << " ";
               if(report.ClassPackets[i] != 0) {
                  pclassvalid[i] = true;
               }
               pclass[i] = pos++;
            }
         }

         if(lipv4) {
            dataos << safeDivision(report.BytesIPv4,report.PacketsIPv4) << " ";
            lipv4 = pos++;
         }
         if(lipv6) {
            dataos << safeDivision(report.BytesIPv6,report.PacketsIPv6) << " ";
            lipv6 = pos++;
         }
         if(lip) {
            dataos << safeDivision(report.BytesIPv4 + report.BytesIPv6,
                                        report.PacketsIPv4 + report.PacketsIPv6) << " ";
            lip = pos++;
         }
         if(ludp) {
            dataos << safeDivision(report.BytesUDP,report.PacketsUDP) << " ";
            ludp = pos++;
         }
         if(ltcp) {
            dataos << safeDivision(report.BytesTCP,report.PacketsTCP) << " ";
            ltcp = pos++;
         }
         if(licmp4) {
            dataos <<  safeDivision(report.BytesICMP4,report.PacketsICMP4) << " ";
            licmp4 = pos++;
         }
         if(licmp6) {
            dataos <<  safeDivision(report.BytesICMP6,report.PacketsICMP6) << " ";
            licmp6 = pos++;
         }
         if(lclasses) {
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               const card64 q = safeDivision(report.ClassBytes[i],report.ClassPackets[i]);
               if(q != 0) {
                  lclassvalid[i] = true;
                }
               dataos << q << " ";
               lclass[i] = pos++;
            }
         }

         // =================================================================

         StreamReport sr;
         while(reader->readNextStreamReport(&sr)) {
            updateStream(sr,time,report.Interval,outputInterval);
         }

         // =================================================================

         dataos << endl;

         if(report.TimeStamp == 0) {
            time += report.Interval;
         }
         else {
            time = report.TimeStamp - tmin;
         }
      }
      else {
         break;
      }
   }

   // ====== Generate GNUplot input =========================================
   gpos << "# Generated by NetAnalyzer - Copyright (C) 2000 Thomas Dreibholz"
        << endl << endl;           
   gpos << "set terminal postscript enhanced color dashed" << endl;
   gpos << "set encoding iso_8859_1" << endl;
   for(cardinal i = 1;i < 65;i++) {
      gpos << "set linestyle " << i << " lt " << i << " lw 5" << endl;
   }

   time_t timeStamp = reader->getTimeStamp() / 1000000;
   const struct tm* timeptr = localtime(&timeStamp);
   char str[256];
   strftime((char*)&str,256,"%A, %B %d %G %T %Z",timeptr);
   String infoString = "Generated on " + (String)str;
   if(info) {
      infoString = infoString + " " + (String)info;
   }

   String timeInfo;
   if(outputInterval == 1000000) {
      timeInfo = " per Second";
   }
   else {
      snprintf((char*)&str,sizeof(str),"%1.6f",outputInterval / 1000000.0);
      for(integer i = strlen(str) - 1;i > 0;i--) {
         if(str[i] == '0') {
            str[i] = 0x00;
         }
         else if(str[i] == '.') {
            str[i] = 0x00;
            break;
         }
         else {
            break;
         }
      }
      timeInfo = String(" within ") + String(str) + " Seconds";
   }
   const String bytesInfo   = String("Bytes ") + timeInfo;
   const String packetsInfo = String("Packets ") + timeInfo;


   if(btcp || btcpdata || budp || budpdata || bicmp4 || bicmp6 || bipv4 || bipv6) {
      const char* title  = "Network Monitor Byte Transmission Statistics";
      const char* xlabel = "Time [s]";
      const char* ylabel = bytesInfo.getData();
      bool firstUsage = true;
      if(btcp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),btcp,"TCP");
      if(btcpdata) plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),btcpdata,"TCP Data");
      if(budp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),budp,"UDP");
      if(budpdata) plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),budpdata,"UDP Data");
      if(bicmp4)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bicmp4,"ICMPv4");
      if(bicmp6)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bicmp6,"ICMPv6");
      if(bipv4)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bipv4,"IPv4");
      if(bipv6)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bipv6,"IPv6");
      if(bip)      plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bip,"IPv4 + IPv6");
      if(!firstUsage) {
         gpos << endl;
         firstUsage = true;
      }
      if(bclasses) {
         for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
            if((bclass[i]) && (bclassvalid[i])) {
               snprintf((char*)&str,sizeof(str),"IPv4/IPv6 Traffic Class %s",TrafficClassValues::getNameForIndex(i));
               plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),bclass[i],(char*)&str);
            }
         }
         gpos << endl;
      }
   }

   if(ptcp || pudp || picmp4 || picmp6 || pipv4 || pipv6) {
      const char* title  = "Network Monitor Packet Transmission Statistics";
      const char* xlabel = "Time [s]";
      const char* ylabel = packetsInfo.getData();
      bool firstUsage = true;
      if(ptcp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),ptcp,"TCP");
      if(pudp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),pudp,"UDP");
      if(picmp4)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),picmp4,"ICMPv4");
      if(picmp6)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),picmp6,"ICMPv6");
      if(pipv4)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),pipv4,"IPv4");
      if(pipv6)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),pipv6,"IPv6");
      if(pip)      plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),pip,"IPv4 + IPv6");
      if(!firstUsage) {
         gpos << endl;
         firstUsage = true;
      }
      if(pclasses) {
         for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
            if((pclass[i]) && (pclassvalid[i])) {
               snprintf((char*)&str,sizeof(str),"IPv4/IPv6 Traffic Class %s",TrafficClassValues::getNameForIndex(i));
               plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),pclass[i],(char*)&str);
            }
         }
         gpos << endl;
      }
   }

   if(ltcp || ludp || licmp4 || licmp6 || lipv4 || lipv6) {
      const char* title  = "Network Monitor Average Packet Length Statistics";
      const char* xlabel = "Time [s]";
      const char* ylabel = "Average Packet Length [Bytes]";
      bool firstUsage = true;
      if(ltcp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),ltcp,"TCP");
      if(ludp)     plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),ludp,"UDP");
      if(licmp4)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),licmp4,"ICMPv4");
      if(licmp6)   plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),licmp6,"ICMPv6");
      if(lipv4)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),lipv4,"IPv4");
      if(lipv6)    plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),lipv6,"IPv6");
      if(lip)      plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),lip,"IPv4 + IPv6");
      if(!firstUsage) {
         gpos << endl;
         firstUsage = true;
      }
      if(lclasses) {
         for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
            if((lclass[i]) && (lclassvalid[i])) {
               snprintf((char*)&str,sizeof(str),"IPv4/IPv6 Traffic Class %s",TrafficClassValues::getNameForIndex(i));
               plot(title,xlabel,ylabel,ownPage,infoString.getData(),gpos,firstUsage,dname.getData(),lclass[i],(char*)&str);
            }
         }
         gpos << endl;
      }
   }

   for(cardinal j = 0;j < StreamCount;j++) {
      InternetAddress source      = InternetAddress(Stream[j]->Source);
      InternetAddress destination = InternetAddress(Stream[j]->Destination);
      source.setPrintFormat((InternetAddress::PrintFormat)format);
      destination.setPrintFormat((InternetAddress::PrintFormat)format);
      char str[32];
      if(Stream[j]->FlowLabel != 0) {
         snprintf((char*)&str,sizeof(str)," / $%05x",Stream[j]->FlowLabel);
      }
      else {
         str[0] = 0x00;
      }
      const String streamName = source.getAddressString() + str +
                                   " -> " + destination.getAddressString();

      if(b1streams) {
         const String title  = streamName + " Raw Transmissions";
         const char* xlabel = "Time [s]";
         const char* ylabel = bytesInfo.getData();
         if(bclasses) {
            bool firstUsage = true;
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               if(Stream[j]->BytesRaw[i] > 0) {
                  String subtitle = (String)"IPv4/IPv6 Traffic Class " + TrafficClassValues::getNameForIndex(i);
                  plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + i,subtitle.getData(),0,maxTime,0,brmax[i]);
               }
            }
            gpos << endl;
         }
         bool firstUsage = true;
         plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + (4 * TrafficClassValues::MaxValues) + 0,"All Traffic Classes",0,maxTime,0,tbrmax);
         gpos << endl;
      }

      if(b2streams) {
         const String title = streamName + " Payload Transmissions";
         const char* xlabel = "Time [s]";
         const char* ylabel = bytesInfo.getData();
         if(bclasses) {
            bool firstUsage = true;
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               if(Stream[j]->BytesPayload[i] > 0) {
                  String subtitle = (String)"IPv4/IPv6 Traffic Class " + TrafficClassValues::getNameForIndex(i);
                  plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + i + (1 * TrafficClassValues::MaxValues),subtitle.getData(),0,maxTime,0,bpmax[i]);
               }
            }
            gpos << endl;
         }
         bool firstUsage = true;
         plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + (4 * TrafficClassValues::MaxValues) + 1,"All Traffic Classes",0,maxTime,0,tbpmax);
         gpos << endl;
      }

      if(pstreams) {
         const String title = streamName + " Packet Transmissions";
         const char* xlabel = "Time [s]";
         const char* ylabel = packetsInfo.getData();
         if(pclasses) {
            bool firstUsage = true;
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               if(Stream[j]->Packets[i] > 0) {
                  String subtitle = (String)"IPv4/IPv6 Traffic Class " + TrafficClassValues::getNameForIndex(i);
                  plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + i + (2 * TrafficClassValues::MaxValues),subtitle.getData(),0,maxTime,0,pmax[i]);
               }
            }
            gpos << endl;
         }
         bool firstUsage = true;
         plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + (4 * TrafficClassValues::MaxValues) + 2,"All Traffic Classes",0,maxTime,0,tpmax);
         gpos << endl;
      }

      if(lstreams) {
         const String title = streamName + " Average Packet Length";
         const char* xlabel = "Time [s]";
         const char* ylabel = "Average Packet Length [Bytes]";
         if(lclasses) {
            bool firstUsage = true;
            for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
               if(Stream[j]->BytesPayload[i] > 0) {
                  String subtitle = (String)"IPv4/IPv6 Traffic Class " + TrafficClassValues::getNameForIndex(i);
                  plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + i + (3 * TrafficClassValues::MaxValues),subtitle.getData(),0,maxTime,0,almax[i]);
               }
            }
            gpos << endl;
         }
         bool firstUsage = true;
         plot(title.getData(),xlabel,ylabel,ownPageStreams,infoString.getData(),gpos,firstUsage,StreamName[j]->getData(),2 + (4 * TrafficClassValues::MaxValues) + 3,"All Traffic Classes",0,maxTime,0,tamax);
         gpos << endl;
      }
   }

   return(0);
}
