// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### NetLog Writer Implementation                                     ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "netlogwriter.h"
#include "timedthread.h"
#include "networkmonitorinterface.h"
#include "streammonitorreport.h"
#include "netlogfileheader.h"
#include "tools.h"


namespace Coral {


// ###### Constructor #######################################################
NetLogWriter::NetLogWriter(const char*              name,
                           const card64             interval,
                           NetworkMonitorInterface* monitor,
                           StreamMonitorReport*     smReport)
   : TimedThread(interval)
{
   // ====== Initialization =================================================
   Monitor        = monitor;
   SMReport       = smReport;
   Count          = 0;
   FileDescriptor = fopen(name,"w");
   if(FileDescriptor == NULL) {
      cerr << "ERROR: NetLogWriter::NetLogWriter() - Unable to write log file <" << name << ">!" << endl;
      return;
   }

   // ====== Write header into file =========================================
   TimeStamp     = getMicroTime();
   Interval      = interval;

   cardinal flags = NLFH_Default;
   if(SMReport != NULL) {
      flags |= NLFH_ContainsStreams;
   }

   NetLogFileHeader nlfh;
   strncpy((char*)&nlfh.ID,NetLogFileHeaderID,4);
   nlfh.Version   = translate16(NetLogFileHeaderVersion);
   nlfh.EntrySize = translate16(sizeof(NetworkMonitorReport));
   nlfh.TimeStamp = translate64(TimeStamp);
   nlfh.Interval  = translate64(Interval);
   nlfh.Flags     = translate32(flags);
   if(fwrite((void *)&nlfh,sizeof(nlfh),1,FileDescriptor) != 1) {
      cerr << "ERROR: NetLogWriter::NetLogWriter() - Write error!" << endl;
      fclose(FileDescriptor);
      FileDescriptor = NULL;
      return;
   }
}


// ###### Destructor ########################################################
NetLogWriter::~NetLogWriter()
{
   stop();
   if(FileDescriptor) {
      fclose(FileDescriptor);
      FileDescriptor = NULL;
   }
}


// ###### TimedThread's timerEvent() implementation #########################
void NetLogWriter::timerEvent()
{
   if(FileDescriptor != NULL) {
      synchronized();

      // Get report from NetworkMonitor
      NetworkMonitorReport report;
      Monitor->lock();
      Monitor->moveReport(&report);

      // Translate report and write it to file
      report.translate();
      if(fwrite((void*)&report,sizeof(report),1,FileDescriptor) != 1) {
         Monitor->unlock();
         cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
         exit(1);
      }

      // card64 t1 = getMicroTime();

      if(SMReport != NULL) {
         const cardinal count = SMReport->getCount();
         card32 data = translate32(count);
         if(fwrite((void*)&data,sizeof(card32),1,FileDescriptor) != 1) {
            Monitor->unlock();
            cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
            exit(1);
         }
         for(cardinal i = 0;i < count;i++) {
            StreamReport* sr = SMReport->getReport(i);
            sr->translate();
            if(fwrite((void*)sr,sizeof(StreamReport),1,FileDescriptor) != 1) {
               Monitor->unlock();
               cerr << "ERROR: NetLogWriter::timerEvent() - Write error!" << endl;
               exit(1);
            }
         }
         SMReport->reset();
      }

      // card64 t2 = getMicroTime();
      // cerr << (t2 - t1) << endl;

      Monitor->unlock();
      Count++;
      unsynchronized();
   }
}


}
