// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Audio Mixer                                                      ####
// ####                                                                  ####
// #### Version 1.00  --  June 22, 2000                                  ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#ifndef AUDIOMIXER_H
#define AUDIOMIXER_H


#include "system.h"

#include <sys/soundcard.h>


namespace RTPAudio {


/**
  * This class is an interface to an audio mixer.
  *
  * @short   Audio Mixer
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class AudioMixer
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     *
     * @param mixerChannel Mixer channel (e.g. SOUND_MIXER_PCM).
     * @param name Mixer device name (e.g. "/dev/mixer").
     */
   AudioMixer(int mixerChannel = SOUND_MIXER_PCM, const char* name = "/dev/mixer");

   /**
     * Destructor.
     */
   ~AudioMixer();


   // ====== Mixer functions ================================================
   /**
     * Check, if mixer is ready.
     *
     * @return true, if mixer is ready; false otherwise.
     */
   inline bool ready() const;

   /**
     * Get volume.
     *
     * @param left Volume of left channel.
     * @param right Volume of right channel.
     * @retrun true, if volume has been written into variables; false otherwise.
     */
   bool getVolume(card8& left, card8& right);

   /**
     * Set volume.
     *
     * @param left Volume of left channel.
     * @param right Volume of right channel.
     * @retrun true, if volume has been set; false otherwise.
     */
   bool setVolume(const card8 left, const card8 right);


   // ====== Private data ===================================================
   private:
   int Device;
   int Channel;
};

}


#endif




// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Audio Mixer implementation                                       ####
// ####                                                                  ####
// #### Version 1.00, March 07, 2000                                     ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
// #include "audiomixer.h"


#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/soundcard.h>


namespace RTPAudio {


// ###### Constructor #######################################################
AudioMixer::AudioMixer(int mixerChannel, const char* name)
{
   Channel = mixerChannel;
   Device = open(name,O_RDWR);
   if(Device < 0) {
      cerr << "WARNING: Unable to open AudioMixer " << name << "." << endl;
   }
}


// ###### Destructor ########################################################
AudioMixer::~AudioMixer()
{
   if(Device >= 0) {
      close(Device);
   }
}


// ###### Get volume ########################################################
bool AudioMixer::getVolume(card8& left, card8& right)
{
   int volume;
   if(ioctl(Device,MIXER_READ(Channel),&volume) >= 0) {
      left = (card8)(volume & 0x7f);
      right = (card8)((volume >> 8) & 0x7f);
      return(true);
   }
   return(false);
}


// ###### Set volume ########################################################
bool AudioMixer::setVolume(const card8 left, const card8 right)
{
   unsigned int l = left;
   unsigned int r = right;
   if(l > 100) l = 100;
   if(r > 100) r = 100;
   int volume = left + (r << 8);
   return(ioctl(Device,MIXER_WRITE(Channel),&volume) >= 0);
}


// ###### Check, if mixer is ready ##########################################
bool AudioMixer::ready() const
{
   return(Device >= 0);
}


}



// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### QAudioMixer                                                      ####
// ####                                                                  ####
// #### Version 1.00, March 07, 2000                                     ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#ifndef QAUDIOMIXER_H
#define QAUDIOMIXER_H


#include "system.h"


#include <qapp.h>
#if QT_VERSION < 200
#error QT Version 2.00 or better required!!!
#endif
#include <qlayout.h>
#include <qpushbutton.h>
#include <qlcdnumber.h>
#include <qlabel.h>
#include <qslider.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qgroupbox.h>
#include <qplatinumstyle.h>
#include <qmainwindow.h>


namespace RTPAudio {


/**
  * This class is a Qt GUI for the audio mixer.
  *
  * @short   QAudioMixer
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class QAudioMixer : public QMainWindow
{
   // ====== Constructor/Destructor =========================================
   Q_OBJECT  
   public:
   /**
     * Constructor.
     *
     * @param mixer AudioMixer object.
     * @param parent Parent widget.
     * @param name Widget name.
     */
   QAudioMixer(AudioMixer* mixer,
               QWidget*    parent = NULL,
               const char* name   = NULL);

   /**
     * Destructor.
     */
   ~QAudioMixer();


   // ====== Qt slots =======================================================
   public slots:
   /**
     * Qt slot: Change balance.
     */
   void balance(int value);

   /**
     * Qt slot: Change volume.
     */
   void volume(int value);

   /**
     * Qt slot: Center balance slider.
     */
   void centerBalance();

   /**
     * Qt slot: Mute.
     */
   void mute();


   // ====== Qt signals =====================================================
   signals:
   /**
     * Qt signal: Emitted, when "Close" or window's close button is clicked.
     */
   void closeAudioMixer();


   // ====== Private data ===================================================
   private:
   void closeEvent(QCloseEvent* event);
   void update();


   int32       VolumeSetting;
   int32       BalanceSetting;

   AudioMixer*  Mixer;
   QPushButton* Mute;
   QSlider*     Balance;
   QSlider*     Volume;
   QLabel*      Values;
   QTimer*      Timer;
};


}


#endif



// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### QAudioMixer implementation                                       ####
// ####                                                                  ####
// #### Version 1.00, March 07, 2000                                     ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####                     EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "wavaudioreader.h"
#include "audiodevice.h"
#include "audiodebug.h"
#include "audioconverter.h"
#include "timedthread.h"
#include "bag.h"


#include <qapp.h>
#if QT_VERSION < 200
#error QT Version 2.00 or better required!!!
#endif
#include <qlayout.h>
#include <qpushbutton.h>
#include <qlcdnumber.h>
#include <qlabel.h>
#include <qslider.h>
#include <qtimer.h>
#include <qpainter.h>
#include <qgroupbox.h>
#include <qplatinumstyle.h>
#include <qmainwindow.h>


// ====== Include Qt's meta objects ======
#include "t5.moc"


namespace RTPAudio {


// ###### Constructor #######################################################
QAudioMixer::QAudioMixer(AudioMixer* mixer,
                         QWidget*    parent,
                         const char* name)
   : QMainWindow(parent,name)
{
   Mixer = mixer;

   // ====== Central widget =================================================
   QWidget* centralWidget = new QWidget(this);
   QVBoxLayout* layout    = new QVBoxLayout(centralWidget);

   // ====== New group ======================================================
   QGroupBox* controlGroup = new QGroupBox("AudioMixer",centralWidget);
   layout->addWidget(controlGroup);
   QGridLayout* controlLayout = new QGridLayout(controlGroup,3,3,20,20);
   controlLayout->setColStretch(1,2);

   // ====== Balance ========================================================
   QLabel* label1 = new QLabel("Balance:",controlGroup);
   controlLayout->addWidget(label1,0,0);
   Balance = new QSlider(0,100,5,50,QSlider::Horizontal,controlGroup);
   Balance->setTickmarks(QSlider::Below);
   Balance->setTickInterval(10);
   controlLayout->addWidget(Balance,0,1);
   QPushButton* center = new QPushButton("Center",controlGroup);
   controlLayout->addWidget(center,0,2);
   QObject::connect(center,SIGNAL(clicked()),this,SLOT(centerBalance()));
   QObject::connect(Balance,SIGNAL(valueChanged(int)),this,SLOT(balance(int)));

   // ====== Volume =========================================================
   QLabel* label2 = new QLabel("Volume:",controlGroup);
   controlLayout->addWidget(label2,1,0);
   Volume = new QSlider(0,100,5,50,QSlider::Horizontal,controlGroup);
   Volume->setTickmarks(QSlider::Below);
   Volume->setTickInterval(10);
   controlLayout->addWidget(Volume,1,1);
   Mute = new QPushButton("Mute",controlGroup);
   Mute->setToggleButton(true);
   controlLayout->addWidget(Mute,1,2);
   QObject::connect(Mute,SIGNAL(clicked()),this,SLOT(mute()));
   QObject::connect(Volume,SIGNAL(valueChanged(int)),this,SLOT(volume(int)));

   // ====== Values =========================================================
   QLabel* label3 = new QLabel("Values:",controlGroup);
   controlLayout->addWidget(label3,2,0);
   Values = new QLabel(controlGroup);
   controlLayout->addMultiCellWidget(Values,2,2,1,2);

   setCentralWidget(centralWidget);


   // ====== Set start values from mixer device =============================
   card8 left,right;
   if(Mixer->getVolume(left,right)) {
      VolumeSetting = (left + right) / 2;
      BalanceSetting = 50 + ((int32)right - (int32)left) / 2;

      Volume->setValue(VolumeSetting);
      Balance->setValue(BalanceSetting);

      char str[32];
      sprintf((char*)&str,"Left %d %%  /  Right %d %%",(card32)left,(card32)right);
      Values->setText((char*)&str);
   }
}


// ###### Destructor ########################################################
QAudioMixer::~QAudioMixer()
{
   delete Timer;
   Timer = NULL;
}


// ###### Update volumes on mixer device ####################################
void QAudioMixer::update()
{
   if(Mute->isOn() == false) {
      const double v = 2.0 * VolumeSetting;
      const double l = (100.0 - BalanceSetting) / 100.0;
      const double r = 1.0 - l;

      const card8 left  = min((card8)VolumeSetting,(card8)(l * v));
      const card8 right = min((card8)VolumeSetting,(card8)(r * v));
      if(Mixer->setVolume(left,right) == false) {
         cerr << "WARNING: QAudioMixer::update() - Error doing AudioMixer::setVolume()!" << endl;
      }

      char str[32];
      sprintf((char*)&str,"Left %d %% / Right %d %%",(card32)left,(card32)right);
      Values->setText((char*)&str);
   }
   else {
      Mixer->setVolume(0,0);
      Values->setText("-- Device Muted --");
   }
}


// ###### Changed balance ###################################################
void QAudioMixer::balance(int value)
{
   BalanceSetting = (int32)value;
   update();
}


// ###### Changed volume ####################################################
void QAudioMixer::volume(int value)
{
   VolumeSetting = (int32)value;
   update();
}


// ###### Mute ##############################################################
void QAudioMixer::mute()
{
   update();
}


// ###### Center balance slider #############################################
void QAudioMixer::centerBalance()
{
   Balance->setValue(50);
}


// ###### Close QAudioMixer #################################################
void QAudioMixer::closeEvent(QCloseEvent* event)
{
   emit closeAudioMixer();
}


}


using namespace RTPAudio;


int main(int argc, char** argv)
{
   QApplication* application = new QApplication(argc,argv);
   application->setStyle(new QPlatinumStyle());

   AudioMixer m;
   QAudioMixer* mixer = new QAudioMixer(&m);

   application->setMainWidget(mixer);
   mixer->show();
   application->exec(); 
}
