// ##########################################################################
// ####                                                                  ####
// ####                      RTP Audio Server Project                    ####
// ####                    ============================                  ####
// ####                                                                  ####
// #### Transport Info Implementation                                    ####
// ####                                                                  ####
// #### Version 1.00  --  February 23, 2001                              ####
// ####                                                                  ####
// #### Copyright (C) 1999  Thomas Dreibholz                             ####
// ####               2000  Universitt Bonn, Abt. IV                    ####
// ####               2001  EMail: Dreibholz@bigfoot.com                 ####
// ####                     WWW:   http://www.bigfoot.com/~dreibholz     ####
// ####                                                                  ####
// ##########################################################################


#include "transportinfo.h"
#include "range.h"
#include "tools.h"


namespace Coral {



// ##########################################################################
// #### TransportInfoLayer                                               ####
// ##########################################################################



// ###### Constructor #######################################################
TransportInfoLayer::TransportInfoLayer()
{
}


// ###### Reset #############################################################
void TransportInfoLayer::reset()
{
   BytesPerSecond   = 0;
   PacketsPerSecond = 0;
   FrameSize        = 0;
   MaxLossRate      = 255;
   Flags            = 0;
   pad              = 0;
}


// ###### Translate byte order ##############################################
void TransportInfoLayer::translate()
{
   BytesPerSecond   = translate64(BytesPerSecond);
   PacketsPerSecond = translate32(PacketsPerSecond);
   FrameSize        = translate32(FrameSize);
   Flags            = translate16(Flags);
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TransportInfoLayer& layer)
{
   char str[32];
   os << "      BytesPerSecond   = " << layer.BytesPerSecond << endl;
   os << "      PacketsPerSecond = " << layer.PacketsPerSecond << endl;
   os << "      FrameSize        = " << layer.FrameSize << endl;
   os << "      MaxLossRate      = " << (cardinal)layer.MaxLossRate << endl;
   snprintf((char*)&str,sizeof(str),"$%04x",layer.Flags);
   os << "      Flags            = " << str << endl;
   return(os);
}



// ##########################################################################
// #### TransportInfoLevel                                               ####
// ##########################################################################



// ###### Constructor #######################################################
TransportInfoLevel::TransportInfoLevel()
{
}


// ###### Reset #############################################################
void TransportInfoLevel::reset()
{
   BytesPerSecondScale   = 65536;
   PacketsPerSecondScale = 65536;
   FramesPerSecond       = 0;
   FramesPerSecondScale  = 65536;
   MaxTransferDelay      = (card32)-1;
   Quality               = 0;
   LevelUp               = (card8)-1;
   LevelDown             = (card8)-1;
   QualityLayers         = 0;
}


// ###### Translate byte order ##############################################
void TransportInfoLevel::translate()
{
   BytesPerSecondScale   = translate32(BytesPerSecondScale);
   PacketsPerSecondScale = translate32(PacketsPerSecondScale);
   FramesPerSecond       = translate32(FramesPerSecond);
   FramesPerSecondScale  = translate32(FramesPerSecondScale);
   MaxTransferDelay      = translate32(MaxTransferDelay);

   for(cardinal i = 0;i < QualityLayers;i++) {
      QualityLayer[i].translate();
   }
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TransportInfoLevel& level)
{
   os << "   BytesPerSecondScale   = " << (level.BytesPerSecondScale / 65536.0) << endl;
   os << "   PacketsPerSecondScale = " << (level.PacketsPerSecondScale / 65536.0) << endl;
   os << "   FramesPerSecond       = " << level.FramesPerSecond << endl;
   os << "   FramesPerSecondScale  = " << (level.FramesPerSecondScale / 65536.0) << endl;
   os << "   MaxTransferDelay      = " << (level.MaxTransferDelay / 16.0) << " [ms]" << endl;
   os << "   Quality               = " << (cardinal)level.Quality << endl;
   os << "   LevelUp               = " << (cardinal)level.LevelUp << endl;
   os << "   LevelDown             = " << (cardinal)level.LevelDown << endl;
   os << "   QualityLayers         = " << (cardinal)level.QualityLayers << endl;
   for(cardinal i = 0;i < min(level.QualityLayers,TransportInfoLevel::MaxQualityLayers);i++) {
      os << "   Layer #" << i << ":" << endl;
      os << level.QualityLayer[i];
   }
   return(os);
}


// ###### Comparision operator ##############################################
int TransportInfoLevel::operator==(const TransportInfoLevel& level) const
{
   if((FramesPerSecondScale != level.FramesPerSecondScale) ||
      (FramesPerSecond      != level.FramesPerSecond)      ||
      (BytesPerSecondScale  != level.BytesPerSecondScale)  ||
      (MaxTransferDelay     != level.MaxTransferDelay)     ||
      (LevelUp              != level.LevelUp)              ||
      (LevelDown            != level.LevelDown)            ||
      (Quality              != level.Quality)              ||
      (QualityLayers        != level.QualityLayers)) {
      return(false);
   }
   for(cardinal i = 0;i < min(QualityLayers,TransportInfoLevel::MaxQualityLayers);i++) {
      if(QualityLayer[i] != level.QualityLayer[i]) {
         return(false);
      }
   }
   return(true);
}




// ##########################################################################
// #### TransportInfo                                                    ####
// ##########################################################################




// ###### Constructor #######################################################
TransportInfo::TransportInfo()
{
}


// ###### Reset #############################################################
void TransportInfo::reset()
{
   for(cardinal i = 0;i < TransportInfoLevel::MaxQualityLayers;i++) {
      WantedBytesPerSecond[i].init(0,0,0);
      WantedPacketsPerSecond[i].init(0,0,0);
      WantedMaxLossRate[i] = 255;
   }
   WantedFramesPerSecond.init(0,0,0);
   WantedMaxTransferDelay = (card32)-1;

   StartFramesPerSecond = 0;
   Flags                = TIF_None;
   QualityLevels        = 0;
   QualityLayers        = 0;

   TotalBytesPerSecondLimit   = 0;
   TotalPacketsPerSecondLimit = 0;
   TotalFramesPerSecondLimit  = 0;

   for(cardinal i = 0;i < MaxQualityLevels;i++) {
      QualityLevel[i].reset();
   }
   CurrentSetting.reset();
}


// ###### Translate byte order ##############################################
void TransportInfo::translate()
{
   for(cardinal i = 0;i < min(QualityLayers,TransportInfoLevel::MaxQualityLayers);i++) {
      WantedBytesPerSecond[i].Min     = translate64(WantedBytesPerSecond[i].Min);
      WantedBytesPerSecond[i].Max     = translate64(WantedBytesPerSecond[i].Max);
      WantedBytesPerSecond[i].Value   = translate64(WantedBytesPerSecond[i].Value);

      WantedPacketsPerSecond[i].Min   = translate32(WantedPacketsPerSecond[i].Min);
      WantedPacketsPerSecond[i].Max   = translate32(WantedPacketsPerSecond[i].Max);
      WantedPacketsPerSecond[i].Value = translate32(WantedPacketsPerSecond[i].Value);
   }

   WantedFramesPerSecond.Min   = translate32(WantedFramesPerSecond.Min);
   WantedFramesPerSecond.Max   = translate32(WantedFramesPerSecond.Max);
   WantedFramesPerSecond.Value = translate32(WantedFramesPerSecond.Value);
   WantedMaxTransferDelay      = translate32(WantedMaxTransferDelay);

   Flags                = translate16(Flags);
   StartFramesPerSecond = translate32(StartFramesPerSecond);

   TotalBytesPerSecondLimit   = translate64(TotalBytesPerSecondLimit);
   TotalPacketsPerSecondLimit = translate32(TotalPacketsPerSecondLimit);
   TotalFramesPerSecondLimit  = translate32(TotalFramesPerSecondLimit);

   for(cardinal i = 0;i < min(QualityLevels,MaxQualityLevels);i++) {
      QualityLevel[i].translate();
   }
   CurrentSetting.translate();
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const TransportInfo& ti)
{
   char str[32];
   os << "Wanted values:" << endl;
   for(cardinal i = 0;i < min(ti.QualityLayers,TransportInfoLevel::MaxQualityLayers);i++) {
      os << "Layer #" << i << ":" << endl;
      os << "   WantedBytesPerSecond   = " << ti.WantedBytesPerSecond[i] << endl;
      os << "   WantedPacketsPerSecond = " << ti.WantedPacketsPerSecond[i] << endl;
      os << "   WantedMaxLossRate      = " << (cardinal)ti.WantedMaxLossRate[i] << endl;
   }
   os << "WantedFramesPerSecond      = " << ti.WantedFramesPerSecond << endl;
   os << "WantedMaxTransferDelay     = " << ti.WantedMaxTransferDelay / 16.0 << " [ms]" << endl;

   snprintf((char*)&str,sizeof(str),"$%04x",ti.Flags);
   os << "Flags                      = " << str << endl;
   os << "StartFramesPerSecond       = " << ti.StartFramesPerSecond << endl;
   os << "QualityLevels              = " << (cardinal)ti.QualityLevels << endl;
   os << "QualityLayers              = " << (cardinal)ti.QualityLayers << endl;
   os << "TotalBytesPerSecondLimit   = " << ti.TotalBytesPerSecondLimit << endl;
   os << "TotalPacketsPerSecondLimit = " << ti.TotalPacketsPerSecondLimit << endl;
   os << "TotalFramesPerSecondLimit  = " << ti.TotalFramesPerSecondLimit << endl;
   
   for(cardinal i = 0;i < min(ti.QualityLevels,TransportInfo::MaxQualityLevels);i++) {
      os << "Level #" << i << ":" << endl;
      os << ti.QualityLevel[i];
   }

   os << "Current Setting:" << endl;
   os << ti.CurrentSetting;
   return(os);
}


// ###### Comparision operator ##############################################
int TransportInfo::operator==(const TransportInfo& ti) const
{
   if((WantedMaxTransferDelay     != ti.WantedMaxTransferDelay)    ||
      (WantedFramesPerSecond      != ti.WantedFramesPerSecond)     ||
      (QualityLayers              != ti.QualityLayers)             ||
      (QualityLevels              != ti.QualityLevels)             ||
      (Flags                      != ti.Flags)                     ||
      (StartFramesPerSecond       != ti.StartFramesPerSecond)      ||
      (TotalBytesPerSecondLimit   != ti.TotalBytesPerSecondLimit)  ||
      (TotalFramesPerSecondLimit  != ti.TotalFramesPerSecondLimit) ||
      (TotalPacketsPerSecondLimit != ti.TotalPacketsPerSecondLimit)) {
      return(false);
   }

   for(cardinal i = 0;i < min(QualityLayers,TransportInfo::MaxQualityLevels);i++) {
      if((WantedBytesPerSecond[i].Min   != ti.WantedBytesPerSecond[i].Min)   ||
         (WantedBytesPerSecond[i].Max   != ti.WantedBytesPerSecond[i].Max)   ||
         (WantedPacketsPerSecond[i].Min != ti.WantedPacketsPerSecond[i].Min) ||
         (WantedPacketsPerSecond[i].Max != ti.WantedPacketsPerSecond[i].Max) ||
         (WantedMaxLossRate[i]          != ti.WantedMaxLossRate[i])) {
         return(false);
      }
   }
   for(cardinal i = 0;i < min(QualityLevels,TransportInfo::MaxQualityLevels);i++) {
      if(QualityLevel[i] != ti.QualityLevel[i]) {
         return(false);
      }
   }
   if(CurrentSetting != ti.CurrentSetting) {
      return(false);
   }
   return(true);
}


}
